import json
from typing import Optional
from unittest import TestCase

from pcs import settings
from pcs.common import reports
from pcs.common.auth import HostAuthData, HostWithTokenAuthData
from pcs.common.communication.const import COM_STATUS_SUCCESS
from pcs.common.communication.dto import InternalCommunicationResultDto
from pcs.common.communication.types import CommunicationResultStatus
from pcs.common.file_type_codes import COROSYNC_CONF, PCS_KNOWN_HOSTS
from pcs.common.host import Destination, PcsKnownHost
from pcs.common.interface.dto import to_dict
from pcs.lib.commands import auth
from pcs.lib.host.config.exporter import Exporter as KnownHostsExporter
from pcs.lib.host.config.types import KnownHosts

from pcs_test.tier0.lib.pcs_cfgsync.test_save_sync import (
    FixtureFetchNewestFileMixin,
)
from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools


def fixture_known_hosts_file_content(
    data_version, hosts: dict[str, PcsKnownHost]
) -> str:
    return KnownHostsExporter.export(
        KnownHosts(
            format_version=1, data_version=data_version, known_hosts=hosts
        )
    ).decode("utf-8")


def fixture_communication_result_string(
    status: CommunicationResultStatus = COM_STATUS_SUCCESS,
    status_msg: Optional[str] = None,
    report_list: Optional[reports.dto.ReportItemDto] = None,
    data="",
) -> str:
    return json.dumps(
        to_dict(
            InternalCommunicationResultDto(
                status=status,
                status_msg=status_msg,
                report_list=report_list or [],
                data=data,
            )
        )
    )


_FIXTURE_KNOWN_HOSTS = {
    "node1": PcsKnownHost("node1", "aaa", [Destination("node1", 2224)]),
    "node2": PcsKnownHost("node2", "bbb", [Destination("node2", 2224)]),
}


class ReadKnownHostsFile(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_file_exists(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS, settings.pcsd_known_hosts_location
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )

        env = self.env_assist.get_env()
        facade = auth._read_known_hosts_file(env.report_processor)
        self.assertEqual(
            facade.config,
            KnownHosts(
                format_version=1,
                data_version=1,
                known_hosts=_FIXTURE_KNOWN_HOSTS,
            ),
        )

    def test_file_does_not_exist(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS, settings.pcsd_known_hosts_location, exists=False
        )

        env = self.env_assist.get_env()
        facade = auth._read_known_hosts_file(env.report_processor)
        self.assertEqual(
            facade.config,
            KnownHosts(format_version=1, data_version=0, known_hosts={}),
        )

    def test_error_reading_file(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS, settings.pcsd_known_hosts_location
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            exception_msg="some error",
        )

        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: auth._read_known_hosts_file(env.report_processor)
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=PCS_KNOWN_HOSTS,
                    operation="read",
                    reason="some error",
                    file_path=settings.pcsd_known_hosts_location,
                )
            ],
        )

    def test_error_reading_file_invalid_file_structure(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS, settings.pcsd_known_hosts_location
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS, settings.pcsd_known_hosts_location, content="A"
        )

        env = self.env_assist.get_env()
        self.env_assist.assert_raise_library_error(
            lambda: auth._read_known_hosts_file(env.report_processor)
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.PARSE_ERROR_JSON_FILE,
                    file_type_code=PCS_KNOWN_HOSTS,
                    line_number=1,
                    column_number=1,
                    position=0,
                    reason="Expecting value",
                    full_msg="Expecting value: line 1 column 1 (char 0)",
                    file_path=settings.pcsd_known_hosts_location,
                )
            ],
        )


class AuthHostsTokenNoSync(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_adds_new(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS, settings.pcsd_known_hosts_location
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.write(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            fixture_known_hosts_file_content(
                2,
                {
                    **_FIXTURE_KNOWN_HOSTS,
                    "node3": PcsKnownHost(
                        "node3", "TOKEN", [Destination("node3", 2224)]
                    ),
                    "node4": PcsKnownHost(
                        "node4", "TOKEN", [Destination("node4", 2224)]
                    ),
                },
            ).encode("utf-8"),
            can_overwrite=True,
        )

        auth.auth_hosts_token_no_sync(
            self.env_assist.get_env(),
            {
                "node3": HostWithTokenAuthData(
                    "TOKEN", [Destination("node3", 2224)]
                ),
                "node4": HostWithTokenAuthData(
                    "TOKEN", [Destination("node4", 2224)]
                ),
            },
        )

    def test_invalid_data_empty_token(self):
        self.env_assist.assert_raise_library_error(
            lambda: auth.auth_hosts_token_no_sync(
                self.env_assist.get_env(),
                {
                    "node2": HostWithTokenAuthData(
                        "", [Destination("node2", 2224)]
                    )
                },
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="token",
                    option_value="",
                    allowed_values="a string (min length: 1) (max length: 512)",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
            ]
        )

    def test_error_writing_file(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS, settings.pcsd_known_hosts_location
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.write(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            fixture_known_hosts_file_content(
                2,
                _FIXTURE_KNOWN_HOSTS
                | {"A": PcsKnownHost("A", "TOKEN", [Destination("A", 2224)])},
            ).encode("utf-8"),
            can_overwrite=True,
            exception_msg="some error",
            exception_action="",
        )

        self.env_assist.assert_raise_library_error(
            lambda: auth.auth_hosts_token_no_sync(
                self.env_assist.get_env(),
                {"A": HostWithTokenAuthData("TOKEN", [Destination("A", 2224)])},
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=PCS_KNOWN_HOSTS,
                    operation="write",
                    reason="some error",
                    file_path=settings.pcsd_known_hosts_location,
                )
            ]
        )


class AuthHosts(TestCase, FixtureFetchNewestFileMixin):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def fixture_send_new_tokens_in_cluster(
        self, new_tokens: dict[str, PcsKnownHost]
    ):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF, settings.corosync_conf_file, name="corosync.exists"
        )
        self.config.env.set_known_nodes(["node1", "node2"])
        self.config.corosync_conf.load(["node1", "node2"])
        self.config.http.pcs_cfgsync.set_configs(
            "test99",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    2, _FIXTURE_KNOWN_HOSTS | new_tokens
                )
            },
            node_labels=["node1", "node2"],
        )

    def test_success_not_in_cluster(self):
        self.config.http.place_multinode_call(
            node_labels=["node3"],
            output="TOKEN",
            action="remote/auth",
            param_list=[("username", "username"), ("password", "password")],
            name="auth",
        )
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF,
            settings.corosync_conf_file,
            exists=False,
            name="corosync.exists",
        )
        self.config.raw_file.write(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            file_data=fixture_known_hosts_file_content(
                2,
                _FIXTURE_KNOWN_HOSTS
                | {
                    "node3": PcsKnownHost(
                        "node3", "TOKEN", [Destination("node3", 2224)]
                    )
                },
            ).encode("utf-8"),
            can_overwrite=True,
        )

        auth.auth_hosts(
            self.env_assist.get_env(),
            {
                "node3": HostAuthData(
                    "username", "password", [Destination("node3", 2224)]
                )
            },
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("node3"),
                ),
            ]
        )

    def test_not_in_cluster_error_writing_file(self):
        self.config.http.place_multinode_call(
            node_labels=["node3"],
            output="TOKEN",
            action="remote/auth",
            param_list=[("username", "username"), ("password", "password")],
            name="auth",
        )
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF,
            settings.corosync_conf_file,
            exists=False,
            name="corosync.exists",
        )
        self.config.raw_file.write(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            file_data=fixture_known_hosts_file_content(
                2,
                _FIXTURE_KNOWN_HOSTS
                | {
                    "node3": PcsKnownHost(
                        "node3", "TOKEN", [Destination("node3", 2224)]
                    )
                },
            ).encode("utf-8"),
            can_overwrite=True,
            exception_msg="some error",
        )

        self.env_assist.assert_raise_library_error(
            lambda: auth.auth_hosts(
                self.env_assist.get_env(),
                {
                    "node3": HostAuthData(
                        "username", "password", [Destination("node3", 2224)]
                    )
                },
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("node3"),
                ),
                fixture.error(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=PCS_KNOWN_HOSTS,
                    operation="write",
                    reason="some error",
                    file_path=settings.pcsd_known_hosts_location,
                ),
            ]
        )

    def test_success_in_cluster(self):
        self.config.http.place_multinode_call(
            node_labels=["node3", "node4"],
            output="TOKEN",
            action="remote/auth",
            param_list=[("username", "username"), ("password", "password")],
            name="auth",
        )
        self.fixture_send_new_tokens_in_cluster(
            {
                "node3": PcsKnownHost(
                    "node3", "TOKEN", [Destination("node3", 2224)]
                ),
                "node4": PcsKnownHost(
                    "node4", "TOKEN", [Destination("node4", 2224)]
                ),
            }
        )

        auth.auth_hosts(
            self.env_assist.get_env(),
            {
                "node3": HostAuthData(
                    "username", "password", [Destination("node3", 2224)]
                ),
                "node4": HostAuthData(
                    "username", "password", [Destination("node4", 2224)]
                ),
            },
        )

        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("node3"),
                ),
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("node4"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto(node="node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto(node="node2"),
                ),
            ]
        )

    def test_success_only_able_to_auth_some_nodes(self):
        self.config.http.place_multinode_call(
            communication_list=[
                {"label": "node3", "output": "TOKEN"},
                {"label": "node4", "output": ""},
            ],
            action="remote/auth",
            param_list=[("username", "username"), ("password", "password")],
            name="auth",
        )
        self.fixture_send_new_tokens_in_cluster(
            {
                "node3": PcsKnownHost(
                    "node3", "TOKEN", [Destination("node3", 2224)]
                )
            }
        )

        self.env_assist.assert_raise_library_error(
            lambda: auth.auth_hosts(
                self.env_assist.get_env(),
                {
                    "node3": HostAuthData(
                        "username", "password", [Destination("node3", 2224)]
                    ),
                    "node4": HostAuthData(
                        "username", "password", [Destination("node4", 2224)]
                    ),
                },
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("node3"),
                ),
                fixture.error(
                    reports.codes.INCORRECT_CREDENTIALS,
                    context=reports.dto.ReportItemContextDto("node4"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto(node="node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto(node="node2"),
                ),
            ]
        )

    def test_unable_to_auth_any_host(self):
        self.config.http.place_multinode_call(
            node_labels=["node3", "node4"],
            output="",  # empty token means that auth was not successful
            action="remote/auth",
            param_list=[("username", "username"), ("password", "password")],
            name="auth",
        )

        self.env_assist.assert_raise_library_error(
            lambda: auth.auth_hosts(
                self.env_assist.get_env(),
                {
                    "node3": HostAuthData(
                        "username", "password", [Destination("node3", 2224)]
                    ),
                    "node4": HostAuthData(
                        "username", "password", [Destination("node4", 2224)]
                    ),
                },
            )
        )

        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INCORRECT_CREDENTIALS,
                    context=reports.dto.ReportItemContextDto("node3"),
                ),
                fixture.error(
                    reports.codes.INCORRECT_CREDENTIALS,
                    context=reports.dto.ReportItemContextDto("node4"),
                ),
            ]
        )

    def test_invalid_auth_data(self):
        self.env_assist.assert_raise_library_error(
            lambda: auth.auth_hosts(
                self.env_assist.get_env(),
                {
                    "node1": HostAuthData("username", "password", []),
                    "node2": HostAuthData(
                        "username", "password", [Destination("", 80)]
                    ),
                },
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="dest_list",
                    option_value="[]",
                    allowed_values="non-empty list of destinations for node 'node1'",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="addr",
                    option_value="",
                    allowed_values="address for node 'node2'",
                    cannot_be_empty=True,
                    forbidden_characters=None,
                ),
            ]
        )

    def test_some_cluster_nodes_not_known_to_pcs(self):
        self.config.http.place_multinode_call(
            node_labels=["node3"],
            output="TOKEN",
            action="remote/auth",
            param_list=[("username", "username"), ("password", "password")],
            name="auth",
        )
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF, settings.corosync_conf_file, name="corosync.exists"
        )
        self.config.env.set_known_nodes(["node1"])
        self.config.corosync_conf.load(["node1", "node2"])
        self.config.http.pcs_cfgsync.set_configs(
            "test99",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    2,
                    _FIXTURE_KNOWN_HOSTS
                    | {
                        "node3": PcsKnownHost(
                            "node3", "TOKEN", [Destination("node3", 2224)]
                        )
                    },
                )
            },
            node_labels=["node1"],
        )
        self.env_assist.assert_raise_library_error(
            lambda: auth.auth_hosts(
                self.env_assist.get_env(),
                {
                    "node3": HostAuthData(
                        "username", "password", [Destination("node3", 2224)]
                    )
                },
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("node3"),
                ),
                fixture.error(
                    reports.codes.HOST_NOT_FOUND, host_list=["node2"]
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto(node="node1"),
                ),
            ]
        )

    def test_no_cluster_nodes_not_known_to_pcs(self):
        self.config.http.place_multinode_call(
            node_labels=["node3"],
            output="TOKEN",
            action="remote/auth",
            param_list=[("username", "username"), ("password", "password")],
            name="auth",
        )
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF, settings.corosync_conf_file, name="corosync.exists"
        )
        self.config.env.set_known_nodes([])
        self.config.corosync_conf.load(["node1", "node2"])

        self.env_assist.assert_raise_library_error(
            lambda: auth.auth_hosts(
                self.env_assist.get_env(),
                {
                    "node3": HostAuthData(
                        "username", "password", [Destination("node3", 2224)]
                    )
                },
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("node3"),
                ),
                fixture.error(
                    reports.codes.HOST_NOT_FOUND, host_list=["node1", "node2"]
                ),
            ]
        )

    def fixture_cfgsync_conflict(self) -> str:
        self.config.http.place_multinode_call(
            node_labels=["NEW"],
            output="TOKEN",
            action="remote/auth",
            param_list=[("username", "username"), ("password", "password")],
            name="auth",
        )
        local_file_version = 1
        local_tokens = {"LOCAL": PcsKnownHost("LOCAL", "LOCAL", [])}

        remote_file_version = 42
        remote_tokens = {"REMOTE": PcsKnownHost("REMOTE", "REMOTE", [])}

        new_tokens = {
            "NEW": PcsKnownHost("NEW", "TOKEN", [Destination("NEW", 2224)]),
        }

        even_more_new_remote_file_version = 69
        even_more_new_remote_tokens = {"WHAT": PcsKnownHost("WHAT", "WHAT", [])}

        local_file = fixture_known_hosts_file_content(
            local_file_version, local_tokens
        )
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=local_file,
        )
        self.config.raw_file.exists(
            COROSYNC_CONF, settings.corosync_conf_file, name="corosync.exists"
        )
        self.config.env.set_known_nodes(["node1", "node2"])
        self.config.corosync_conf.load(["node1", "node2"])
        self.config.http.pcs_cfgsync.set_configs(
            "test99",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    local_file_version + 1, local_tokens | new_tokens
                )
            },
            communication_list=[
                {
                    "label": "node1",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "rejected"},
                        }
                    ),
                },
                {"label": "node2"},
            ],
            name="set_configs.1",
        )

        # fetching the newest config from cluster
        remote_file = fixture_known_hosts_file_content(
            remote_file_version, remote_tokens
        )
        self.fixture_fetch_newest_file(
            local_file, remote_file, call_name_suffix="1", cluster_name="test99"
        )

        # sending the merged tokens
        self.config.http.pcs_cfgsync.set_configs(
            "test99",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    remote_file_version + 1,
                    local_tokens | new_tokens | remote_tokens,
                )
            },
            communication_list=[
                {
                    "label": "node1",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "rejected"},
                        }
                    ),
                },
                {"label": "node2"},
            ],
            name="set_configs.2",
        )

        # fetching the even more newest config from cluster
        even_more_new_remote_file = fixture_known_hosts_file_content(
            even_more_new_remote_file_version, even_more_new_remote_tokens
        )
        self.fixture_fetch_newest_file(
            local_file,
            even_more_new_remote_file,
            call_name_suffix="2",
            cluster_name="test99",
        )

        return even_more_new_remote_file

    def test_conflict_upon_conflict(self):
        file_to_save = self.fixture_cfgsync_conflict()
        # write the new file
        self.config.raw_file.write(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            file_to_save.encode("utf-8"),
            can_overwrite=True,
        )

        self.env_assist.assert_raise_library_error(
            lambda: auth.auth_hosts(
                self.env_assist.get_env(),
                {
                    "NEW": HostAuthData(
                        "username", "password", [Destination("NEW", 2224)]
                    )
                },
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("NEW"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.error(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.error(reports.codes.PCS_CFGSYNC_CONFLICT_REPEAT_ACTION),
            ]
        )

    def test_conflict_upon_conflict_error_saving_new_file(self):
        file_to_save = self.fixture_cfgsync_conflict()
        # write the new file
        self.config.raw_file.write(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            file_to_save.encode("utf-8"),
            can_overwrite=True,
            exception_msg="error",
        )

        self.env_assist.assert_raise_library_error(
            lambda: auth.auth_hosts(
                self.env_assist.get_env(),
                {
                    "NEW": HostAuthData(
                        "username", "password", [Destination("NEW", 2224)]
                    )
                },
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.AUTHORIZATION_SUCCESSFUL,
                    context=reports.dto.ReportItemContextDto("NEW"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.error(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.error(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=PCS_KNOWN_HOSTS,
                    operation="write",
                    reason="error",
                    file_path=settings.pcsd_known_hosts_location,
                ),
                fixture.error(reports.codes.PCS_CFGSYNC_CONFLICT_REPEAT_ACTION),
            ]
        )


class DeauthHosts(TestCase, FixtureFetchNewestFileMixin):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_success_not_in_cluster(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF,
            settings.corosync_conf_file,
            exists=False,
            name="corosync.exists",
        )
        self.config.raw_file.write(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            file_data=fixture_known_hosts_file_content(
                2, {"node1": _FIXTURE_KNOWN_HOSTS["node1"]}
            ).encode("utf-8"),
            can_overwrite=True,
        )

        auth.deauth_hosts(self.env_assist.get_env(), ["node2"])

    def test_not_in_cluster_error_writing_file(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF,
            settings.corosync_conf_file,
            exists=False,
            name="corosync.exists",
        )
        self.config.raw_file.write(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            file_data=fixture_known_hosts_file_content(
                2, {"node1": _FIXTURE_KNOWN_HOSTS["node1"]}
            ).encode("utf-8"),
            can_overwrite=True,
            exception_msg="some error",
        )

        self.env_assist.assert_raise_library_error(
            lambda: auth.deauth_hosts(self.env_assist.get_env(), ["node2"])
        )

        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=PCS_KNOWN_HOSTS,
                    operation="write",
                    reason="some error",
                    file_path=settings.pcsd_known_hosts_location,
                ),
            ]
        )

    def test_success_in_cluster(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF, settings.corosync_conf_file, name="corosync.exists"
        )
        self.config.env.set_known_nodes(["node1", "node2"])
        self.config.corosync_conf.load(["node1", "node2"])
        self.config.http.pcs_cfgsync.set_configs(
            "test99",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    2, {"node1": _FIXTURE_KNOWN_HOSTS["node1"]}
                )
            },
            node_labels=["node1", "node2"],
        )

        auth.deauth_hosts(self.env_assist.get_env(), ["node2"])
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto(node="node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto(node="node2"),
                ),
            ]
        )

    def test_some_hosts_not_found(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF, settings.corosync_conf_file, name="corosync.exists"
        )
        self.config.env.set_known_nodes(["node1", "node2"])
        self.config.corosync_conf.load(["node1", "node2"])
        self.config.http.pcs_cfgsync.set_configs(
            "test99",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    2, {"node1": _FIXTURE_KNOWN_HOSTS["node1"]}
                )
            },
            node_labels=["node1", "node2"],
        )

        self.env_assist.assert_raise_library_error(
            lambda: auth.deauth_hosts(
                self.env_assist.get_env(), ["node2", "node3"]
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.HOST_NOT_FOUND, host_list=["node3"]
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto(node="node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto(node="node2"),
                ),
            ]
        )

    def test_no_hosts_found(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )

        self.env_assist.assert_raise_library_error(
            lambda: auth.deauth_hosts(
                self.env_assist.get_env(), ["node3", "node4"]
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.HOST_NOT_FOUND, host_list=["node3", "node4"]
                )
            ]
        )

    def test_no_hosts_specified(self):
        self.env_assist.assert_raise_library_error(
            lambda: auth.deauth_hosts(self.env_assist.get_env(), [])
        )
        self.env_assist.assert_reports(
            [fixture.error(reports.codes.NO_HOST_SPECIFIED)]
        )

    def test_some_cluster_nodes_not_known_to_pcs(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF, settings.corosync_conf_file, name="corosync.exists"
        )
        self.config.env.set_known_nodes(["node1"])
        self.config.corosync_conf.load(["node1", "node2"])
        self.config.http.pcs_cfgsync.set_configs(
            "test99",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    2, {"node1": _FIXTURE_KNOWN_HOSTS["node1"]}
                )
            },
            node_labels=["node1"],
        )

        self.env_assist.assert_raise_library_error(
            lambda: auth.deauth_hosts(self.env_assist.get_env(), ["node2"])
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.HOST_NOT_FOUND, host_list=["node2"]
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto(node="node1"),
                ),
            ]
        )

    def test_no_cluster_nodes_not_known_to_pcs(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF, settings.corosync_conf_file, name="corosync.exists"
        )
        self.config.env.set_known_nodes([])
        self.config.corosync_conf.load(["node1", "node2"])

        self.env_assist.assert_raise_library_error(
            lambda: auth.deauth_hosts(self.env_assist.get_env(), ["node2"])
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.HOST_NOT_FOUND, host_list=["node1", "node2"]
                ),
                fixture.error(reports.codes.NONE_HOST_FOUND),
            ]
        )

    def fixture_cfgsync_conflict(self):
        local_file_version = 1
        local_tokens = {"LOCAL": PcsKnownHost("LOCAL", "LOCAL", [])}

        remote_file_version = 42
        remote_tokens = {"REMOTE": PcsKnownHost("REMOTE", "REMOTE", [])}

        even_more_new_remote_file_version = 69
        even_more_new_remote_tokens = {"WHAT": PcsKnownHost("WHAT", "WHAT", [])}

        local_file = fixture_known_hosts_file_content(
            local_file_version, local_tokens
        )
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=local_file,
        )
        self.config.raw_file.exists(
            COROSYNC_CONF, settings.corosync_conf_file, name="corosync.exists"
        )
        self.config.env.set_known_nodes(["node1", "node2"])
        self.config.corosync_conf.load(["node1", "node2"])
        self.config.http.pcs_cfgsync.set_configs(
            "test99",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    local_file_version + 1, {}
                )
            },
            communication_list=[
                {
                    "label": "node1",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "rejected"},
                        }
                    ),
                },
                {"label": "node2"},
            ],
            name="set_configs.1",
        )

        # fetching the newest config from cluster
        remote_file = fixture_known_hosts_file_content(
            remote_file_version, remote_tokens
        )
        self.fixture_fetch_newest_file(
            local_file, remote_file, call_name_suffix="1", cluster_name="test99"
        )

        # sending the merged tokens
        self.config.http.pcs_cfgsync.set_configs(
            "test99",
            {
                PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(
                    remote_file_version + 1,
                    remote_tokens,
                )
            },
            communication_list=[
                {
                    "label": "node1",
                    "output": json.dumps(
                        {
                            "status": "ok",
                            "result": {PCS_KNOWN_HOSTS: "rejected"},
                        }
                    ),
                },
                {"label": "node2"},
            ],
            name="set_configs.2",
        )

        # fetching the even more newest config from cluster
        even_more_new_remote_file = fixture_known_hosts_file_content(
            even_more_new_remote_file_version, even_more_new_remote_tokens
        )
        self.fixture_fetch_newest_file(
            local_file,
            even_more_new_remote_file,
            call_name_suffix="2",
            cluster_name="test99",
        )

        return even_more_new_remote_file

    def test_conflict_upon_conflict(self):
        file_to_save = self.fixture_cfgsync_conflict()
        # write the new file
        self.config.raw_file.write(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            file_to_save.encode("utf-8"),
            can_overwrite=True,
        )

        self.env_assist.assert_raise_library_error(
            lambda: auth.deauth_hosts(self.env_assist.get_env(), ["LOCAL"])
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.error(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.error(reports.codes.PCS_CFGSYNC_CONFLICT_REPEAT_ACTION),
            ]
        )

    def test_conflict_upon_conflict_error_saving_new_file(self):
        file_to_save = self.fixture_cfgsync_conflict()
        # write the new file
        self.config.raw_file.write(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            file_to_save.encode("utf-8"),
            can_overwrite=True,
            exception_msg="error",
        )

        self.env_assist.assert_raise_library_error(
            lambda: auth.deauth_hosts(self.env_assist.get_env(), ["LOCAL"])
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.error(
                    reports.codes.PCS_CFGSYNC_CONFIG_REJECTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto("node2"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_FETCHING_NEWEST_CONFIG,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.error(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=PCS_KNOWN_HOSTS,
                    operation="write",
                    reason="error",
                    file_path=settings.pcsd_known_hosts_location,
                ),
                fixture.error(reports.codes.PCS_CFGSYNC_CONFLICT_REPEAT_ACTION),
            ]
        )


class DeauthAllLocalHosts(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_success_not_in_cluster(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF,
            settings.corosync_conf_file,
            exists=False,
            name="corosync.exists",
        )
        self.config.raw_file.write(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            file_data=fixture_known_hosts_file_content(2, {}).encode("utf-8"),
            can_overwrite=True,
        )

        auth.deauth_all_local_hosts(self.env_assist.get_env())

    def test_not_in_cluster_error_writing_file(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF,
            settings.corosync_conf_file,
            exists=False,
            name="corosync.exists",
        )
        self.config.raw_file.write(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            file_data=fixture_known_hosts_file_content(2, {}).encode("utf-8"),
            can_overwrite=True,
            exception_msg="some error",
        )

        self.env_assist.assert_raise_library_error(
            lambda: auth.deauth_all_local_hosts(self.env_assist.get_env())
        )

        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=PCS_KNOWN_HOSTS,
                    operation="write",
                    reason="some error",
                    file_path=settings.pcsd_known_hosts_location,
                ),
            ]
        )

    def test_success_in_cluster(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, _FIXTURE_KNOWN_HOSTS),
        )
        self.config.raw_file.exists(
            COROSYNC_CONF, settings.corosync_conf_file, name="corosync.exists"
        )
        self.config.env.set_known_nodes(["node1", "node2"])
        self.config.corosync_conf.load(["node1", "node2"])
        self.config.http.pcs_cfgsync.set_configs(
            "test99",
            {PCS_KNOWN_HOSTS: fixture_known_hosts_file_content(2, {})},
            node_labels=["node1", "node2"],
        )

        auth.deauth_all_local_hosts(self.env_assist.get_env())
        self.env_assist.assert_reports(
            [
                fixture.info(
                    reports.codes.PCS_CFGSYNC_SENDING_CONFIGS_TO_NODES,
                    file_type_code_list=[PCS_KNOWN_HOSTS],
                    node_name_list=["node1", "node2"],
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto(node="node1"),
                ),
                fixture.info(
                    reports.codes.PCS_CFGSYNC_CONFIG_ACCEPTED,
                    file_type_code=PCS_KNOWN_HOSTS,
                    context=reports.dto.ReportItemContextDto(node="node2"),
                ),
            ]
        )

    def test_no_hosts_in_local_file(self):
        self.config.raw_file.exists(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.exists",
        )
        self.config.raw_file.read(
            PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_file_content(1, {}),
        )

        auth.deauth_all_local_hosts(self.env_assist.get_env())



from dataclasses import dataclass
from typing import (
    Optional,
    Sequence,
    Union,
)

from pcs.common.const import (
    PcmkRoleType,
    PcmkStatusRoleType,
)
from pcs.common.interface.dto import DataTransferObject


@dataclass(frozen=True)
class PrimitiveStatusDto(DataTransferObject):
    # pylint: disable=too-many-instance-attributes
    resource_id: str
    instance_id: Optional[str]
    resource_agent: str
    role: PcmkStatusRoleType
    target_role: Optional[PcmkRoleType]
    active: bool
    orphaned: bool
    blocked: bool
    maintenance: bool
    description: Optional[str]
    failed: bool
    managed: bool
    failure_ignored: bool
    node_names: list[str]
    pending: Optional[str]
    locked_to: Optional[str]


@dataclass(frozen=True)
class GroupStatusDto(DataTransferObject):
    resource_id: str
    instance_id: Optional[str]
    maintenance: bool
    description: Optional[str]
    managed: bool
    disabled: bool
    members: Sequence[PrimitiveStatusDto]


@dataclass(frozen=True)
class CloneStatusDto(DataTransferObject):
    # pylint: disable=too-many-instance-attributes
    resource_id: str
    multi_state: bool
    unique: bool
    maintenance: bool
    description: Optional[str]
    managed: bool
    disabled: bool
    failed: bool
    failure_ignored: bool
    target_role: Optional[PcmkRoleType]
    instances: Union[Sequence[PrimitiveStatusDto], Sequence[GroupStatusDto]]


@dataclass(frozen=True)
class BundleReplicaStatusDto(DataTransferObject):
    replica_id: str
    member: Optional[PrimitiveStatusDto]
    remote: Optional[PrimitiveStatusDto]
    container: PrimitiveStatusDto
    ip_address: Optional[PrimitiveStatusDto]


@dataclass(frozen=True)
class BundleStatusDto(DataTransferObject):
    # pylint: disable=too-many-instance-attributes
    resource_id: str
    type: str
    image: str
    unique: bool
    maintenance: bool
    description: Optional[str]
    managed: bool
    failed: bool
    replicas: Sequence[BundleReplicaStatusDto]


AnyResourceStatusDto = Union[
    PrimitiveStatusDto, GroupStatusDto, CloneStatusDto, BundleStatusDto
]


@dataclass(frozen=True)
class ResourcesStatusDto(DataTransferObject):
    resources: Sequence[AnyResourceStatusDto]



from collections.abc import Sequence

from pcs.common.pacemaker.fencing_topology import (
    CibFencingLevel,
    CibFencingLevelAttributeDto,
    CibFencingLevelRegexDto,
    CibFencingTopologyDto,
)
from pcs.common.str_tools import indent
from pcs.common.types import (
    StringCollection,
    StringSequence,
)


def _get_targets_with_levels_str(
    levels: Sequence[CibFencingLevel],
) -> list[str]:
    lines = []
    last_target_value = ""
    for level in levels:
        if isinstance(level, CibFencingLevelAttributeDto):
            target_label = "attribute"
            target_value = f"{level.target_attribute}={level.target_value}"
        elif isinstance(level, CibFencingLevelRegexDto):
            target_label = "regexp"
            target_value = level.target_pattern
        else:
            target_label = "node"
            target_value = level.target
        if target_value != last_target_value:
            lines.append(f"Target ({target_label}): {target_value}")
        last_target_value = target_value
        lines.extend(
            indent(
                [
                    "Level {level}: {devices}".format(
                        level=level.index, devices=" ".join(level.devices)
                    )
                ]
            )
        )
    return lines


def stonith_level_config_to_text(
    fencing_topology: CibFencingTopologyDto,
) -> StringSequence:
    target_node_levels = sorted(
        fencing_topology.target_node,
        key=lambda level: (level.target, level.index),
    )
    target_regex_levels = sorted(
        fencing_topology.target_regex,
        key=lambda level: (level.target_pattern, level.index),
    )
    target_attr_levels = sorted(
        fencing_topology.target_attribute,
        key=lambda level: (
            level.target_value,
            level.target_attribute,
            level.index,
        ),
    )

    return (
        _get_targets_with_levels_str(target_node_levels)
        + _get_targets_with_levels_str(target_regex_levels)
        + _get_targets_with_levels_str(target_attr_levels)
    )


def _get_level_add_cmd(
    index: int,
    target: str,
    device_list: StringCollection,
    level_id: str,
) -> str:
    devices = " ".join(device_list)
    return (
        f"pcs stonith level add --force -- {index} {target} {devices} "
        f"id={level_id}"
    )


def stonith_level_config_to_cmd(
    fencing_topology: CibFencingTopologyDto,
) -> StringSequence:
    lines: list[str] = []
    lines.extend(
        _get_level_add_cmd(level.index, level.target, level.devices, level.id)
        for level in fencing_topology.target_node
    )
    for level_regex in fencing_topology.target_regex:
        target = f"regexp%{level_regex.target_pattern}"
        lines.append(
            _get_level_add_cmd(
                level_regex.index, target, level_regex.devices, level_regex.id
            )
        )
    for level_attr in fencing_topology.target_attribute:
        target = (
            f"attrib%{level_attr.target_attribute}={level_attr.target_value}"
        )
        lines.append(
            _get_level_add_cmd(
                level_attr.index, target, level_attr.devices, level_attr.id
            )
        )

    return lines



from unittest import TestCase

from pcs.common import reports
from pcs.lib.commands import resource

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools


class ResourceRestart(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        resources_xml = """
            <resources>
                <primitive id="S1" class="stonith" />
                <primitive id="R1" class="ocf">
                    <meta_attributes id="R1-meta" />
                </primitive>
                <bundle id="B1">
                    <primitive id="B1R1" />
                </bundle>
                <clone id="C1">
                    <primitive id="C1R1" />
                </clone>
                <clone id="C2">
                    <group id="C2G1">
                        <primitive id="C2R1" />
                    </group>
                </clone>
            </resources>
        """
        self.config.runner.cib.load(resources=resources_xml)

    def test_success(self):
        self.config.runner.pcmk.resource_restart("R1")
        resource.restart(self.env_assist.get_env(), "R1")

    def test_success_bundle_member(self):
        self.config.runner.pcmk.resource_restart("B1")
        resource.restart(self.env_assist.get_env(), "B1R1")
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.RESOURCE_RESTART_USING_PARENT_RESOURCE,
                    resource="B1R1",
                    parent="B1",
                )
            ]
        )

    def test_success_clone_member(self):
        self.config.runner.pcmk.resource_restart("C1")
        resource.restart(self.env_assist.get_env(), "C1R1")
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.RESOURCE_RESTART_USING_PARENT_RESOURCE,
                    resource="C1R1",
                    parent="C1",
                )
            ]
        )

    def test_success_clone_group_member(self):
        self.config.runner.pcmk.resource_restart("C2")
        resource.restart(self.env_assist.get_env(), "C2R1")
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.RESOURCE_RESTART_USING_PARENT_RESOURCE,
                    resource="C2R1",
                    parent="C2",
                )
            ]
        )

    def test_success_timeout(self):
        self.config.runner.pcmk.resource_restart("R1", timeout="10")
        resource.restart(self.env_assist.get_env(), "R1", timeout="10")

    def test_success_all_options(self):
        self.config.runner.pcmk.resource_restart(
            "C1", node="node1", timeout="10"
        )
        resource.restart(self.env_assist.get_env(), "C1", "node1", "10")

    def test_bad_timeout(self):
        self.env_assist.assert_raise_library_error(
            lambda: resource.restart(
                self.env_assist.get_env(), "R1", timeout="a while"
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="timeout",
                    option_value="a while",
                    allowed_values="time interval (e.g. 1, 2s, 3m, 4h, ...)",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ]
        )

    def test_resource_not_found(self):
        self.config.runner.pcmk.resource_restart("RX")
        resource.restart(self.env_assist.get_env(), "RX")

    def test_not_a_resource(self):
        self.env_assist.assert_raise_library_error(
            lambda: resource.restart(self.env_assist.get_env(), "R1-meta")
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="R1-meta",
                    expected_types=["resource"],
                    current_type="meta_attributes",
                )
            ]
        )

    def test_stonith(self):
        self.env_assist.assert_raise_library_error(
            lambda: resource.restart(self.env_assist.get_env(), "S1")
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.COMMAND_ARGUMENT_TYPE_MISMATCH,
                    not_accepted_type="stonith resource",
                    command_to_use_instead=None,
                )
            ]
        )

    def test_node_not_multitinstance(self):
        self.env_assist.assert_raise_library_error(
            lambda: resource.restart(
                self.env_assist.get_env(), "R1", node="node1"
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_RESTART_NODE_IS_FOR_MULTIINSTANCE_ONLY,
                    resource="R1",
                    resource_type="primitive",
                    node="node1",
                )
            ]
        )

    def test_node_multitinstance(self):
        self.config.runner.pcmk.resource_restart("C1", node="node1")
        resource.restart(self.env_assist.get_env(), "C1", "node1")

    def test_restart_error(self):
        self.config.runner.pcmk.resource_restart(
            "R1", stdout="some output", stderr="some error", returncode=1
        )
        self.env_assist.assert_raise_library_error(
            lambda: resource.restart(self.env_assist.get_env(), "R1"),
            [
                fixture.error(
                    reports.codes.RESOURCE_RESTART_ERROR,
                    reason="some error\nsome output",
                    resource="R1",
                    node=None,
                )
            ],
            expected_in_processor=False,
        )



import json
from typing import Iterable

from tornado.web import HTTPError

from pcs.common.capabilities import Capability
from pcs.daemon.app.auth import (
    LegacyTokenAuthenticationHandler,
    NotAuthorizedException,
    TokenAuthProvider,
)
from pcs.lib.auth.provider import AuthProvider
from pcs.lib.auth.types import AuthUser

from .common import (
    BaseHandler,
    RoutesType,
)


def _capabilities_to_str(capabilities: Iterable[Capability]) -> str:
    return json.dumps(
        {
            "pcsd_capabilities": [feat.code for feat in capabilities],
        }
    )


class LegacyCapabilitiesHandler(LegacyTokenAuthenticationHandler):
    _capabilities: Iterable[Capability]

    def initialize(
        self, auth_provider: AuthProvider, capabilities: Iterable[Capability]
    ) -> None:
        # pylint: disable=arguments-differ
        super().initialize(auth_provider)
        self._capabilities = capabilities

    async def _handle_request(self) -> None:
        self.write(_capabilities_to_str(self._capabilities))


class CapabilitiesHandler(BaseHandler):
    _auth_provider: TokenAuthProvider
    _capabilities: Iterable[Capability]

    def initialize(
        self, auth_provider: AuthProvider, capabilities: Iterable[Capability]
    ) -> None:
        super().initialize()
        self._auth_provider = TokenAuthProvider(self, auth_provider)
        self._capabilities = capabilities

    async def get_auth_user(self) -> AuthUser:
        try:
            return await self._auth_provider.auth_by_token()
        except NotAuthorizedException as e:
            raise HTTPError(401) from e

    async def get(self) -> None:
        await self.get_auth_user()
        self.add_header("Content-Type", "application/json")
        self.write(_capabilities_to_str(self._capabilities))


def get_routes(
    auth_provider: AuthProvider, capabilities: Iterable[Capability]
) -> RoutesType:
    """
    Returns mapping of URL routes to functions
    """
    params = dict(auth_provider=auth_provider, capabilities=capabilities)
    return [
        ("/remote/capabilities", LegacyCapabilitiesHandler, params),
        ("/capabilities", CapabilitiesHandler, params),
    ]



from dataclasses import dataclass
from itertools import chain
from typing import Sequence

from pcs.common.interface.dto import DataTransferObject

from .bundle import CibResourceBundleDto
from .clone import CibResourceCloneDto
from .group import CibResourceGroupDto
from .primitive import CibResourcePrimitiveDto


@dataclass(frozen=True)
class CibResourcesDto(DataTransferObject):
    primitives: Sequence[CibResourcePrimitiveDto]
    clones: Sequence[CibResourceCloneDto]
    groups: Sequence[CibResourceGroupDto]
    bundles: Sequence[CibResourceBundleDto]


def get_all_resources_ids(resources_dto: CibResourcesDto) -> set[str]:
    return set(
        chain(
            (primitive.id for primitive in resources_dto.primitives),
            (group.id for group in resources_dto.groups),
            (clone.id for clone in resources_dto.clones),
            (bundle.id for bundle in resources_dto.bundles),
        )
    )


def get_stonith_resources_ids(resources_dto: CibResourcesDto) -> set[str]:
    return {
        primitive.id
        for primitive in resources_dto.primitives
        if primitive.agent_name.standard == "stonith"
    }



from dataclasses import dataclass
from typing import Optional, Sequence

from pcs.common.interface.dto import DataTransferObject
from pcs.common.pacemaker.nvset import CibNvsetDto


@dataclass(frozen=True)
class CibAlertRecipientDto(DataTransferObject):
    id: str
    value: str
    description: Optional[str]
    meta_attributes: Sequence[CibNvsetDto]
    instance_attributes: Sequence[CibNvsetDto]


@dataclass(frozen=True)
class CibAlertSelectAttributeDto(DataTransferObject):
    id: str
    name: str


@dataclass(frozen=True)
class CibAlertSelectDto(DataTransferObject):
    nodes: bool
    fencing: bool
    resources: bool
    attributes: bool
    attributes_select: Sequence[CibAlertSelectAttributeDto]


@dataclass(frozen=True)
class CibAlertDto(DataTransferObject):
    id: str
    path: str
    description: Optional[str]
    recipients: Sequence[CibAlertRecipientDto]
    select: Optional[CibAlertSelectDto]
    meta_attributes: Sequence[CibNvsetDto]
    instance_attributes: Sequence[CibNvsetDto]


@dataclass(frozen=True)
class CibAlertListDto(DataTransferObject):
    alerts: Sequence[CibAlertDto]



from dataclasses import dataclass
from typing import (
    Any,
    Mapping,
)

from pcs import settings


@dataclass(frozen=True)
class Destination:
    addr: str
    port: int


@dataclass(frozen=True)
class PcsKnownHost:
    name: str
    token: str
    dest_list: list[Destination]

    @classmethod
    def from_known_host_file_dict(
        cls, name: str, known_host_dict: Mapping[str, Any]
    ) -> "PcsKnownHost":
        dest_list = [
            Destination(conn["addr"], conn["port"])
            for conn in known_host_dict["dest_list"]
        ]
        if not dest_list:
            raise KeyError("no destination defined")
        return cls(name, token=known_host_dict["token"], dest_list=dest_list)

    def to_known_host_dict(self) -> tuple[str, dict[str, Any]]:
        return (
            self.name,
            dict(
                token=self.token,
                dest_list=[
                    dict(
                        addr=dest.addr,
                        port=dest.port,
                    )
                    for dest in self.dest_list
                ],
            ),
        )

    @property
    def dest(self) -> Destination:
        if self.dest_list:
            return self.dest_list[0]
        return Destination(self.name, settings.pcsd_default_port)



from typing import Optional
from unittest import TestCase, mock

from pcs.cli.resource import command
from pcs.common import const
from pcs.common.const import PcmkRoleType, PcmkStatusRoleType
from pcs.common.status_dto import (
    BundleReplicaStatusDto,
    BundleStatusDto,
    CloneStatusDto,
    GroupStatusDto,
    PrimitiveStatusDto,
    ResourcesStatusDto,
)


def _fixture_primitive_status_dto(
    resource_id: str,
    resource_agent: str = "ocf:pacemaker:Dummy",
    target_role: Optional[PcmkRoleType] = None,
    role: PcmkStatusRoleType = const.PCMK_STATUS_ROLE_STARTED,
    managed: bool = True,
    node_names: Optional[list[str]] = None,
) -> PrimitiveStatusDto:
    return PrimitiveStatusDto(
        resource_id=resource_id,
        instance_id=None,
        resource_agent=resource_agent,
        role=role,
        target_role=target_role,
        active=False,
        orphaned=False,
        blocked=False,
        maintenance=False,
        description=None,
        failed=False,
        managed=managed,
        failure_ignored=False,
        node_names=node_names if node_names else [],
        pending=None,
        locked_to=None,
    )


def _fixture_group_status_dto(
    group_size,
    running_on_nodes,
    instance_id=None,
    agent_name="ocf:pacemaker:Stateful",
):
    return GroupStatusDto(
        resource_id="G1",
        instance_id=instance_id,
        maintenance=False,
        description=None,
        managed=True,
        disabled=False,
        members=[
            _fixture_primitive_status_dto(
                f"R{i}",
                agent_name,
                node_names=running_on_nodes,
            )
            for i in range(1, group_size)
        ],
    )


def _fixture_bundle_status_dto(replicas):
    return BundleStatusDto(
        resource_id="B1",
        type="docker",
        image="pcs:test",
        unique=True,
        maintenance=False,
        description=None,
        managed=False,
        failed=False,
        replicas=replicas,
    )


def _fixture_bundle_replica_dto(replica_id, container):
    return BundleReplicaStatusDto(
        replica_id=replica_id,
        member=None,
        remote=None,
        ip_address=None,
        container=container,
    )


def _fixture_clone_status_dto(multi_state, instances):
    return CloneStatusDto(
        resource_id="clone",
        multi_state=multi_state,
        unique=False,
        maintenance=False,
        description=None,
        managed=True,
        disabled=False,
        failed=False,
        failure_ignored=False,
        target_role=None,
        instances=instances,
    )


class GetResourceMessage(TestCase):
    def setUp(self):
        self.lib = mock.Mock(spec_set=["status"])

        self.lib.status = mock.Mock(spec_set=["resources_status"])
        self.lib.status.resources_status = self.mock_resources_status = (
            mock.Mock()
        )

    def test_bundle_not_running(self):
        self.mock_resources_status.return_value = ResourcesStatusDto(
            [
                _fixture_bundle_status_dto(
                    replicas=[
                        _fixture_bundle_replica_dto(
                            replica_id=0,
                            container=_fixture_primitive_status_dto(
                                "B1-docker-0",
                                "ocf:heartbeat:docker",
                                role=const.PCMK_STATUS_ROLE_STOPPED,
                                managed=False,
                                node_names=["node1"],
                            ),
                        ),
                    ],
                )
            ]
        )
        self.assertEqual(
            command.get_resource_status_msg(self.lib, "B1"),
            "Resource 'B1' is not running on any nodes",
        )

    def test_bundle_running_one_replica(self):
        self.mock_resources_status.return_value = ResourcesStatusDto(
            [
                _fixture_bundle_status_dto(
                    replicas=[
                        _fixture_bundle_replica_dto(
                            replica_id="0",
                            container=_fixture_primitive_status_dto(
                                "B1-docker-0",
                                "ocf:heartbeat:docker",
                                node_names=["node1"],
                            ),
                        ),
                    ],
                )
            ]
        )
        self.assertEqual(
            command.get_resource_status_msg(self.lib, "B1"),
            "Resource 'B1' is running on node 'node1'",
        )

    def test_bundle_running_and_failed(self):
        self.mock_resources_status.return_value = ResourcesStatusDto(
            [
                _fixture_bundle_status_dto(
                    replicas=[
                        _fixture_bundle_replica_dto(
                            replica_id="0",
                            container=_fixture_primitive_status_dto(
                                "B1-docker-0",
                                "ocf:heartbeat:docker",
                                node_names=["node1"],
                            ),
                        ),
                        _fixture_bundle_replica_dto(
                            replica_id="1",
                            container=_fixture_primitive_status_dto(
                                "B1-docker-1",
                                "ocf:heartbeat:docker",
                                node_names=["node2"],
                            ),
                        ),
                        _fixture_bundle_replica_dto(
                            replica_id="2",
                            container=_fixture_primitive_status_dto(
                                "B1-docker-2",
                                "ocf:heartbeat:docker",
                                role=const.PCMK_STATUS_ROLE_STOPPED,
                                node_names=[],
                            ),
                        ),
                    ],
                )
            ]
        )
        self.assertEqual(
            command.get_resource_status_msg(self.lib, "B1"),
            "Resource 'B1' is running on nodes 'node1', 'node2'",
        )

    def test_bundle_running_multiple_replicas_one_node(self):
        self.mock_resources_status.return_value = ResourcesStatusDto(
            [
                _fixture_bundle_status_dto(
                    replicas=[
                        _fixture_bundle_replica_dto(
                            replica_id="0",
                            container=_fixture_primitive_status_dto(
                                "B1-docker-0",
                                "ocf:heartbeat:docker",
                                node_names=["node1"],
                            ),
                        ),
                        _fixture_bundle_replica_dto(
                            replica_id="1",
                            container=_fixture_primitive_status_dto(
                                "B1-docker-1",
                                "ocf:heartbeat:docker",
                                node_names=["node1"],
                            ),
                        ),
                    ],
                )
            ]
        )
        self.assertEqual(
            command.get_resource_status_msg(self.lib, "B1"),
            "Resource 'B1' is running on node 'node1'",
        )

    def test_primitive(self):
        status_expected_msg = {
            const.PCMK_STATUS_ROLE_STARTED: "Resource 'R1' is running on node 'node1'"
        }
        status_expected_msg.update(
            dict.fromkeys(
                [
                    const.PCMK_STATUS_ROLE_STARTING,
                    const.PCMK_STATUS_ROLE_STOPPING,
                    const.PCMK_STATUS_ROLE_STOPPED,
                ],
                "Resource 'R1' is not running on any nodes",
            )
        )
        for status, expected_msg in status_expected_msg.items():
            with self.subTest(status=status, expected_msg=expected_msg):
                self.mock_resources_status.return_value = ResourcesStatusDto(
                    [
                        _fixture_primitive_status_dto(
                            "R1",
                            role=status,
                            node_names=["node1"],
                        ),
                    ]
                )
                self.assertEqual(
                    command.get_resource_status_msg(self.lib, "R1"),
                    expected_msg,
                )

    def test_cloned_group_one_instance(self):
        self.mock_resources_status.return_value = ResourcesStatusDto(
            [
                _fixture_clone_status_dto(
                    multi_state=True,
                    instances=[
                        _fixture_group_status_dto(
                            instance_id=0,
                            group_size=3,
                            running_on_nodes=["node1"],
                        )
                    ],
                ),
            ]
        )
        self.assertEqual(
            command.get_resource_status_msg(self.lib, "clone"),
            "Resource 'clone' is running on node 'node1'",
        )

    def test_cloned_group_more_instances(self):
        self.mock_resources_status.return_value = ResourcesStatusDto(
            [
                _fixture_clone_status_dto(
                    multi_state=True,
                    instances=[
                        _fixture_group_status_dto(
                            instance_id=0,
                            group_size=3,
                            running_on_nodes=["node1"],
                        ),
                        _fixture_group_status_dto(
                            instance_id=1,
                            group_size=3,
                            running_on_nodes=["node2"],
                        ),
                    ],
                ),
            ]
        )
        self.assertEqual(
            command.get_resource_status_msg(self.lib, "clone"),
            "Resource 'clone' is running on nodes 'node1', 'node2'",
        )

    def test_clone_one_instance(self):
        self.mock_resources_status.return_value = ResourcesStatusDto(
            [
                _fixture_clone_status_dto(
                    multi_state=False,
                    instances=[
                        _fixture_primitive_status_dto(
                            "clone",
                            node_names=["node1"],
                        )
                    ],
                ),
            ]
        )
        self.assertEqual(
            command.get_resource_status_msg(self.lib, "clone"),
            "Resource 'clone' is running on node 'node1'",
        )

    def test_clone_more_instances(self):
        self.mock_resources_status.return_value = ResourcesStatusDto(
            [
                _fixture_clone_status_dto(
                    multi_state=False,
                    instances=[
                        _fixture_primitive_status_dto(
                            "R6", node_names=["node1"]
                        ),
                        _fixture_primitive_status_dto(
                            "R6", node_names=["node1"]
                        ),
                        _fixture_primitive_status_dto(
                            "R6", node_names=["node2"]
                        ),
                        _fixture_primitive_status_dto(
                            "R6", node_names=["node2"]
                        ),
                    ],
                ),
            ]
        )
        self.assertEqual(
            command.get_resource_status_msg(self.lib, "clone"),
            "Resource 'clone' is running on nodes 'node1', 'node2'",
        )

    def test_clone_promotable(self):
        self.mock_resources_status.return_value = ResourcesStatusDto(
            [
                _fixture_clone_status_dto(
                    multi_state=False,
                    instances=[
                        _fixture_primitive_status_dto(
                            "R6",
                            node_names=["node1"],
                            role=const.PCMK_STATUS_ROLE_PROMOTED,
                        ),
                        _fixture_primitive_status_dto(
                            "R6",
                            node_names=["node2"],
                            role=const.PCMK_STATUS_ROLE_UNPROMOTED,
                        ),
                        _fixture_primitive_status_dto(
                            "R6",
                            node_names=["node3"],
                            role=const.PCMK_STATUS_ROLE_DEMOTING,
                        ),
                        _fixture_primitive_status_dto(
                            "R6",
                            node_names=["node3"],
                            role=const.PCMK_STATUS_ROLE_PROMOTING,
                        ),
                    ],
                ),
            ]
        )
        self.assertEqual(
            command.get_resource_status_msg(self.lib, "clone"),
            "Resource 'clone' is promoted on node 'node1'; unpromoted on "
            "node 'node2'",
        )

    def test_group(self):
        self.mock_resources_status.return_value = ResourcesStatusDto(
            [
                _fixture_group_status_dto(
                    group_size=2,
                    running_on_nodes=["node1", "node2"],
                    agent_name="ocf:pacemaker:Dummy",
                ),
            ]
        )
        self.assertEqual(
            command.get_resource_status_msg(self.lib, "G1"),
            "Resource 'G1' is running on nodes 'node1', 'node2'",
        )



from dataclasses import dataclass
from textwrap import dedent
from typing import Iterable

from lxml import etree

from pcs import settings
from pcs.common.tools import xml_fromstring


@dataclass(frozen=True)
class Capability:
    code: str
    description: str
    in_pcs: bool
    in_pcsd: bool


class CapabilitiesError(Exception):
    def __init__(self, msg: str):
        super().__init__()
        self.msg = msg

    def __str__(self) -> str:
        return self.msg


def get_capabilities_definition() -> list[Capability]:
    """
    Read and parse capabilities file

    The point is to return all data in python structures for further processing.
    """
    filename = settings.pcs_capabilities
    try:
        with open(filename, mode="r") as xml_file:
            capabilities_xml = xml_fromstring(xml_file.read())
    except (OSError, etree.XMLSyntaxError, etree.DocumentInvalid) as e:
        raise CapabilitiesError(
            f"Cannot read capabilities definition file '{filename}': '{e}'"
        ) from e
    capabilities = []
    for feat_xml in capabilities_xml.findall(".//capability"):
        desc_elem = feat_xml.find("./description")
        # dedent and strip remove indentation in the XML file
        desc = "" if desc_elem is None else dedent(desc_elem.text or "").strip()
        capabilities.append(
            Capability(
                code=str(feat_xml.attrib["id"]),
                description=desc,
                in_pcs=feat_xml.attrib["in-pcs"] == "1",
                in_pcsd=feat_xml.attrib["in-pcsd"] == "1",
            )
        )
    return capabilities


def get_pcs_capabilities() -> list[Capability]:
    """
    Get pcs capabilities from the capabilities file
    """
    return [feat for feat in get_capabilities_definition() if feat.in_pcs]


def get_pcsd_capabilities() -> list[Capability]:
    """
    Get pcsd capabilities from the capabilities file
    """
    return [feat for feat in get_capabilities_definition() if feat.in_pcsd]


def capabilities_to_codes_str(capabilities: Iterable[Capability]) -> str:
    return " ".join(sorted([feat.code for feat in capabilities]))



# Exporting rule xml elements to strings is an integral part of exporting rule
# xml elements to DTOs. There are extensive tests for that in
# pcs_test/tier0/lib/cib/rule/test_cib_to_dto.py and other tests in
# pcs_test/tier0/lib/cib/test_nvpair_multi.py and
# pcs_test/tier0/lib/commands/test_cib_options.py.
# Therefore we don't duplicate those here. However, if there's a need to write
# specific tests here, feel free to do so.


from unittest import TestCase

from lxml import etree

from pcs.lib.cib.rule.cib_to_str import RuleToStr


class IsoToStr(TestCase):
    # pylint: disable=protected-access
    def test_no_change(self):
        self.assertEqual(RuleToStr._date_to_str("2023-06"), "2023-06")
        self.assertEqual(RuleToStr._date_to_str("202306"), "202306")
        self.assertEqual(RuleToStr._date_to_str("2023-06-30"), "2023-06-30")
        self.assertEqual(RuleToStr._date_to_str("20230630"), "20230630")
        self.assertEqual(
            RuleToStr._date_to_str("2023-06-30T16:30"), "2023-06-30T16:30"
        )
        self.assertEqual(
            RuleToStr._date_to_str("20230630T1630"), "20230630T1630"
        )
        self.assertEqual(
            RuleToStr._date_to_str("2023-06-30T16:30Z"), "2023-06-30T16:30Z"
        )
        self.assertEqual(
            RuleToStr._date_to_str("20230630T1630+2"), "20230630T1630+2"
        )
        self.assertEqual(
            RuleToStr._date_to_str("2023-06-30T16:30:40+2:00"),
            "2023-06-30T16:30:40+2:00",
        )
        self.assertEqual(
            RuleToStr._date_to_str("20230630T1630+02:00"), "20230630T1630+02:00"
        )

    def test_remove_spaces(self):
        self.assertEqual(RuleToStr._date_to_str("- 2023"), "-2023")
        self.assertEqual(RuleToStr._date_to_str("+ 2023"), "+2023")
        self.assertEqual(RuleToStr._date_to_str("2023- 06"), "2023-06")
        self.assertEqual(RuleToStr._date_to_str("2023 -06- 30"), "2023-06-30")
        self.assertEqual(
            RuleToStr._date_to_str("2023-06-30  T16:30"), "2023-06-30T16:30"
        )
        self.assertEqual(
            RuleToStr._date_to_str("20230630T   1630"), "20230630T1630"
        )
        self.assertEqual(
            RuleToStr._date_to_str("2023-06-30 T 16:30  Z"), "2023-06-30T16:30Z"
        )
        self.assertEqual(
            RuleToStr._date_to_str("20230630  T 1630 + 2"), "20230630T1630+2"
        )
        self.assertEqual(
            RuleToStr._date_to_str(
                "2023 -  06 - 30  T 16 :  30 :  40 +  2: 00"
            ),
            "2023-06-30T16:30:40+2:00",
        )
        self.assertEqual(
            RuleToStr._date_to_str("20230630  T   1630+ 02:00"),
            "20230630T1630+02:00",
        )

    def test_add_time_separator(self):
        self.assertEqual(
            RuleToStr._date_to_str("2023-06-30  16:30"), "2023-06-30T16:30"
        )
        self.assertEqual(
            RuleToStr._date_to_str("20230630 1630"), "20230630T1630"
        )
        self.assertEqual(
            RuleToStr._date_to_str("2023-06-30 16:30  Z"), "2023-06-30T16:30Z"
        )
        self.assertEqual(
            RuleToStr._date_to_str("20230630   1630 + 2"), "20230630T1630+2"
        )
        self.assertEqual(
            RuleToStr._date_to_str("2023 -  06 - 30   16 :  30 :  40 +  2: 00"),
            "2023-06-30T16:30:40+2:00",
        )
        self.assertEqual(
            RuleToStr._date_to_str("20230630     1630+ 02:00"),
            "20230630T1630+02:00",
        )

    def test_extra_spaces(self):
        self.assertEqual(
            RuleToStr._date_to_str("2023-06-30 16:30:40 +2 00"),
            "2023-06-30T16:30:40+2 00",
        )
        self.assertEqual(
            RuleToStr._date_to_str("2023 06 30 16 30 +02"),
            "2023T06 30 16 30+02",
        )


class NormalizedStr(TestCase):
    def test_success(self):
        xml = etree.fromstring(
            """
            <rule boolean-op="or" id="R1" score="INFINITY">
              <rule id="R1-rule-1" boolean-op="and" score="0">
                <date_expression id="R1-rule-1-expr" operation="date_spec">
                  <date_spec id="R1-rule-1-expr-datespec"
                      weekdays="1-5" hours="12-23"
                  />
                </date_expression>
                <date_expression id="R1-rule-1-expr-1"
                    operation="in_range" start="2014-07-26"
                >
                  <duration id="R1-rule-1-expr-1-duration" months="1"/>
                </date_expression>
              </rule>
              <rule id="R1-rule" boolean-op="and" score="0">
                <expression id="R1-rule-expr-1"
                    attribute="foo" operation="gt" type="version" value="1.2"
                />
                <expression id="R1-rule-expr"
                    attribute="#uname" operation="eq" value="node3 4"
                />
              </rule>
            </rule>
            """
        )
        self.assertEqual(
            RuleToStr(normalize=True).get_str(xml),
            (
                '(#uname eq string "node3 4" and foo gt version 1.2) or '
                "(date in_range 2014-07-26 to duration months=1 and "
                "date-spec hours=12-23 weekdays=1-5)"
            ),
        )



import json
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from typing import Any

from dacite import DaciteError

from pcs.common import reports
from pcs.common.communication.const import (
    COM_STATUS_SUCCESS,
    COM_STATUS_UNKNOWN_CMD,
)
from pcs.common.communication.dto import InternalCommunicationResultDto
from pcs.common.file_type_codes import (
    PCS_KNOWN_HOSTS,
    PCS_SETTINGS_CONF,
    FileTypeCode,
)
from pcs.common.interface.dto import PayloadConversionError, from_dict
from pcs.common.node_communicator import (
    Request,
    RequestData,
    RequestTarget,
    Response,
)
from pcs.common.pcs_cfgsync_dto import SyncConfigsDto
from pcs.common.reports import ReportProcessor
from pcs.common.reports.processor import has_errors
from pcs.lib.communication.tools import (
    AllAtOnceStrategyMixin,
    AllSameDataMixin,
    RunRemotelyBase,
    SkipOfflineMixin,
)

_LEGACY_FILE_NAME_TO_FILETYPECODE_MAP = {
    "known-hosts": PCS_KNOWN_HOSTS,
    "pcs_settings.conf": PCS_SETTINGS_CONF,
}
_FILETYPECODE_TO_LEGACY_FILE_NAME_MAP = {
    PCS_KNOWN_HOSTS: "known-hosts",
    PCS_SETTINGS_CONF: "pcs_settings.conf",
}


@dataclass(frozen=True)
class ConfigInfo:
    cfg_origin: str
    cfg_content: str


@dataclass(frozen=True)
class GetConfigsResult:
    was_successful: bool
    config_files: dict[FileTypeCode, list[ConfigInfo]]


class GetConfigs(
    SkipOfflineMixin,
    AllSameDataMixin,
    AllAtOnceStrategyMixin,
    RunRemotelyBase,
):
    _LEGACY_ENDPOINT = "remote/get_configs"

    def __init__(
        self,
        report_processor: ReportProcessor,
        cluster_name: str,
        skip_offline_targets: bool = False,
    ):
        super().__init__(report_processor)
        self._cluster_name = cluster_name
        self._successful_connections = 0
        self._received_configs: dict[FileTypeCode, list[ConfigInfo]] = (
            defaultdict(list)
        )
        self._set_skip_offline(skip_offline_targets)

    def _get_request_data(self) -> RequestData:
        return RequestData(
            "api/v1/cfgsync-get-configs/v1",
            data=json.dumps({"cluster_name": self._cluster_name}),
        )

    def _get_legacy_request(self, target: RequestTarget) -> Request:
        return Request(
            target,
            RequestData(
                self._LEGACY_ENDPOINT, [("cluster_name", self._cluster_name)]
            ),
        )

    def _process_response(self, response: Response) -> list[Request]:  # noqa: PLR0911
        if response.request.action == self._LEGACY_ENDPOINT:
            self._process_legacy_response(response)
            return []

        request_target = response.request.target

        if response.response_code == 404:
            # If we communicate with older node, that does not support
            # the new endpoint, then try using the old endpoint
            return [self._get_legacy_request(request_target)]

        report_item = self._get_response_report(response)
        if report_item:
            self._report(report_item)
            return []

        try:
            com_result: InternalCommunicationResultDto = from_dict(
                InternalCommunicationResultDto, json.loads(response.data)
            )
        except (json.JSONDecodeError, DaciteError):
            self._report(
                reports.ReportItem.error(
                    reports.messages.InvalidResponseFormat(request_target.label)
                )
            )
            return []

        if com_result.status == COM_STATUS_UNKNOWN_CMD:
            # If we communicate with older node, that does not support
            # the new endpoint, then try using the old one
            return [self._get_legacy_request(request_target)]

        context = reports.ReportItemContext(response.request.target.label)
        report_list = [
            reports.report_dto_to_item(report, context)
            for report in com_result.report_list
        ]
        self._report_list(report_list)
        errors_in_report_list = has_errors(report_list)

        if (
            not errors_in_report_list
            and com_result.status == COM_STATUS_SUCCESS
        ):
            try:
                config_data: SyncConfigsDto = from_dict(
                    SyncConfigsDto, com_result.data
                )
            except (DaciteError, PayloadConversionError):
                self._report(
                    reports.ReportItem.error(
                        reports.messages.InvalidResponseFormat(
                            request_target.label
                        )
                    )
                )
                return []
            if config_data.cluster_name != self._cluster_name:
                self._report(
                    reports.ReportItem.error(
                        reports.messages.NodeReportsUnexpectedClusterName(
                            self._cluster_name
                        ),
                        context=context,
                    )
                )
                return []

            self._successful_connections += 1
            for cfg_type, cfg_content in config_data.configs.items():
                self._received_configs[cfg_type].append(
                    ConfigInfo(request_target.label, cfg_content)
                )
            return []

        # Make sure we report an error when the command was not successful
        if com_result.status_msg or not errors_in_report_list:
            self._report(
                reports.ReportItem.error(
                    reports.messages.NodeCommunicationCommandUnsuccessful(
                        request_target.label,
                        response.request.action,
                        com_result.status_msg or "Unknown error",
                    )
                )
            )
        return []

    def _process_legacy_response(self, response: Response) -> None:
        """
        format of the `response.data`:
        {
            "status": str -> ("ok"|"not_in_cluster"|"wrong_cluster_name")
            "cluster_name": str
            "configs": {
                "pcs_settings.conf": {
                    "type": str -> ("file"),
                    "text": Optional[str]
                },
                "known-hosts": {
                    "type": str -> ("file"),
                    "text": Optional[str]
                }
            }
        }
        """
        report_item = self._get_response_report(response)
        if report_item:
            self._report(report_item)
            return
        context = reports.ReportItemContext(response.request.target.label)

        try:
            parsed_data = json.loads(response.data)
            if (
                parsed_data["status"] == "wrong_cluster_name"
                or parsed_data["status"] == "not_in_cluster"
                or parsed_data["cluster_name"] != self._cluster_name
            ):
                self._report(
                    reports.ReportItem.error(
                        reports.messages.NodeReportsUnexpectedClusterName(
                            self._cluster_name
                        ),
                        context=context,
                    )
                )
                return

            for cfg_name in parsed_data["configs"]:
                if cfg_name not in _LEGACY_FILE_NAME_TO_FILETYPECODE_MAP:
                    continue
                cfg_file = parsed_data["configs"][cfg_name]["text"]
                if (
                    parsed_data["configs"][cfg_name]["type"] == "file"
                    and cfg_file is not None
                ):
                    self._received_configs[
                        _LEGACY_FILE_NAME_TO_FILETYPECODE_MAP[cfg_name]
                    ].append(
                        ConfigInfo(response.request.target.label, cfg_file)
                    )

            self._successful_connections += 1
        except (json.JSONDecodeError, KeyError):
            self._report(
                reports.ReportItem.error(
                    reports.messages.InvalidResponseFormat(
                        response.request.target.label
                    )
                )
            )

    def on_complete(self) -> GetConfigsResult:
        return GetConfigsResult(
            was_successful=self._successful_connections >= 2,
            config_files=self._received_configs,
        )


class SetConfigsResult(Enum):
    ACCEPTED = "accepted"
    REJECTED = "rejected"
    NOT_SUPPORTED = "not_supported"
    ERROR = "error"


class SetConfigs(
    AllSameDataMixin,
    AllAtOnceStrategyMixin,
    RunRemotelyBase,
):
    def __init__(
        self,
        report_processor: ReportProcessor,
        cluster_name: str,
        configs: dict[FileTypeCode, str],
        rejection_severity: reports.ReportItemSeverity,
        force: bool = False,
    ):
        super().__init__(report_processor)
        self._cluster_name = cluster_name
        self._configs = configs
        self._rejection_severity = rejection_severity
        self._force = force
        self._node_results: dict[str, dict[FileTypeCode, SetConfigsResult]] = {}

    def before(self) -> None:
        self._report(
            reports.ReportItem.info(
                reports.messages.PcsCfgsyncSendingConfigsToNodes(
                    sorted(self._configs.keys()), self._target_label_list
                )
            )
        )

    def _get_request_data(self) -> RequestData:
        data: dict[str, Any] = {}
        data["cluster_name"] = self._cluster_name
        data["force"] = self._force
        data["configs"] = {
            _FILETYPECODE_TO_LEGACY_FILE_NAME_MAP.get(
                filetype_code, filetype_code
            ): {
                "type": "file",
                "text": config_content,
            }
            for filetype_code, config_content in self._configs.items()
        }
        return RequestData(
            "remote/set_configs", [("configs", json.dumps(data))]
        )

    def _process_response(self, response: Response) -> list[Request]:
        report_item = self._get_response_report(response)
        if report_item:
            self._report(report_item)
            return []

        node_label = response.request.target.label
        context = reports.ReportItemContext(node_label)

        try:
            parsed_data = json.loads(response.data)
            if parsed_data["status"] == "wrong_cluster_name":
                self._report(
                    reports.ReportItem.error(
                        reports.messages.NodeReportsUnexpectedClusterName(
                            self._cluster_name or ""
                        ),
                        context=context,
                    )
                )
                return []

            parsed_result = parsed_data["result"]
            results = {}
            for cfg_name in parsed_result:
                cfg_filetypecode = _LEGACY_FILE_NAME_TO_FILETYPECODE_MAP.get(
                    cfg_name, cfg_name
                )
                normalized_cfg_result = SetConfigsResult(
                    parsed_result[cfg_name]
                )
                match normalized_cfg_result:
                    case SetConfigsResult.ACCEPTED:
                        self._report(
                            reports.ReportItem.info(
                                reports.messages.PcsCfgsyncConfigAccepted(
                                    cfg_filetypecode
                                ),
                                context=reports.ReportItemContext(node_label),
                            )
                        )
                    case SetConfigsResult.REJECTED:
                        self._report(
                            reports.ReportItem(
                                self._rejection_severity,
                                reports.messages.PcsCfgsyncConfigRejected(
                                    cfg_filetypecode
                                ),
                                context=reports.ReportItemContext(node_label),
                            )
                        )
                    case SetConfigsResult.ERROR:
                        # The api does not provide description of the error, so we
                        # just provide some generic error message
                        self._report(
                            reports.ReportItem.error(
                                reports.messages.PcsCfgsyncConfigSaveError(
                                    cfg_filetypecode,
                                ),
                                context=reports.ReportItemContext(node_label),
                            )
                        )
                    case SetConfigsResult.NOT_SUPPORTED:
                        self._report(
                            reports.ReportItem.error(
                                reports.messages.PcsCfgsyncConfigUnsupported(
                                    cfg_filetypecode
                                ),
                                context=reports.ReportItemContext(node_label),
                            )
                        )

                results[cfg_filetypecode] = normalized_cfg_result
            self._node_results[node_label] = results
        except (json.JSONDecodeError, KeyError, ValueError, AttributeError):
            self._report(
                reports.ReportItem.error(
                    reports.messages.InvalidResponseFormat(node_label)
                )
            )

        return []

    def on_complete(self) -> dict[str, dict[FileTypeCode, SetConfigsResult]]:
        return self._node_results



from typing import (
    Any,
    cast,
)

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    Argv,
    InputModifiers,
)
from pcs.cli.constraint import command
from pcs.cli.constraint.output import print_config
from pcs.cli.reports.output import deprecation_warning
from pcs.cli.reports.preprocessor import (
    get_duplicate_constraint_exists_preprocessor,
)
from pcs.common.pacemaker.constraint import CibConstraintsDto


def create_with_set(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    create colocation constraint with resource set
    object lib exposes library
    list argv see usage for "constraint colocation set"
    dict like object modifiers can contain
        "force" allows resource in clone/master and constraint duplicity

    Options:
      * --force - allow resource inside clone (or master), allow duplicate
        element
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f", "--force")
    lib.env.report_processor.set_report_item_preprocessor(
        get_duplicate_constraint_exists_preprocessor(lib)
    )
    command.create_with_set(
        lib.constraint_colocation.create_with_set,
        argv,
        modifiers,
    )


def config_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    modifiers.ensure_only_supported("-f", "--output-format", "--full")
    if argv:
        raise CmdLineInputError()

    constraints_dto = cast(
        CibConstraintsDto,
        lib.constraint.get_config(evaluate_rules=True),
    )

    print_config(
        CibConstraintsDto(
            colocation=constraints_dto.colocation,
            colocation_set=constraints_dto.colocation_set,
        ),
        modifiers,
    )


def remove(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -f - CIB file
    """
    modifiers.ensure_only_supported("-f")
    if len(argv) != 2:
        raise CmdLineInputError()
    source_rsc_id, target_rsc_id = argv
    constraint_ids_to_remove: list[str] = []
    constraint_dto = lib.constraint.get_config(evaluate_rules=False)
    for colocation_dto in constraint_dto.colocation:
        if (
            colocation_dto.resource_id == source_rsc_id
            and colocation_dto.with_resource_id == target_rsc_id
        ):
            constraint_ids_to_remove.append(
                colocation_dto.attributes.constraint_id
            )
        elif (
            colocation_dto.resource_id == target_rsc_id
            and colocation_dto.with_resource_id == source_rsc_id
        ):
            # deprecated since pcs-0.11.7
            deprecation_warning(
                "Removing colocation constraint with interchanged source "
                "resource id and target resource id. This behavior is "
                "deprecated and will be removed."
            )
            constraint_ids_to_remove.append(
                colocation_dto.attributes.constraint_id
            )
    if not constraint_ids_to_remove:
        raise CmdLineInputError(
            f"Unable to find colocation constraint with source resource "
            f"'{source_rsc_id}' and target resource '{target_rsc_id}'"
        )
    lib.cib.remove_elements(constraint_ids_to_remove)



from typing import Any, Mapping, Sequence, cast

from pcs.common import reports
from pcs.lib import validate
from pcs.lib.cib.constraint import common, ticket
from pcs.lib.cib.tools import (
    ElementNotFound,
    IdProvider,
    get_constraints,
    get_element_by_id,
    get_pacemaker_version_by_which_cib_was_validated,
)
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError

from .common import _load_resource_set_list, _primitive_resource_set_list


def create(
    env: LibraryEnvironment,
    ticket_key: str,
    resource_id: str,
    options: Mapping[str, str],
    resource_in_clone_alowed: bool = False,
    duplication_alowed: bool = False,
) -> None:
    """
    create a plain ticket constraint

    ticket_key -- ticket for constraining a resource
    resource_id -- resource to be constrained
    options -- desired constraint attributes
    resource_in_clone_alowed -- allow to constrain a resource in a clone
    duplication_alowed -- allow to create a duplicate constraint
    """
    cib = env.get_cib()
    id_provider = IdProvider(cib)
    constraint_section = get_constraints(cib)

    # validation
    constrained_el = None
    try:
        constrained_el = get_element_by_id(cib, resource_id)
    except ElementNotFound:
        env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.IdNotFound(resource_id, [])
            )
        )

    options_pairs = validate.values_to_pairs(
        options,
        validate.option_value_normalization(
            {
                "loss-policy": lambda value: value.lower(),
                "rsc-role": lambda value: value.capitalize(),
            }
        ),
    )

    env.report_processor.report_list(
        ticket.validate_create_plain(
            id_provider,
            ticket_key,
            constrained_el,
            options_pairs,
            in_multiinstance_allowed=resource_in_clone_alowed,
        )
    )

    if env.report_processor.has_errors:
        raise LibraryError()

    # modify CIB
    new_constraint = ticket.create_plain(
        constraint_section,
        id_provider,
        get_pacemaker_version_by_which_cib_was_validated(cib),
        ticket_key,
        resource_id,
        validate.pairs_to_values(options_pairs),
    )

    # Check whether the created constraint is a duplicate of an existing one
    env.report_processor.report_list(
        ticket.DuplicatesCheckerTicketPlain().check(
            constraint_section,
            new_constraint,
            {reports.codes.FORCE} if duplication_alowed else set(),
        )
    )
    if env.report_processor.has_errors:
        raise LibraryError()

    # push CIB
    env.push_cib()


def create_with_set(
    env: LibraryEnvironment,
    resource_set_list: Sequence[Mapping[str, Any]],
    constraint_options: Mapping[str, str],
    resource_in_clone_alowed: bool = False,
    duplication_alowed: bool = False,
) -> None:
    """
    create a set ticket constraint

    resource_set_list -- description of resource sets, for example:
        {"ids": ["A", "B"], "options": {"sequential": "true"}},
    constraint_options -- desired constraint attributes
    resource_in_clone_alowed -- allow to constrain resources in a clone
    duplication_alowed -- allow to create a duplicate constraint
    """
    cib = env.get_cib()
    id_provider = IdProvider(cib)
    constraint_section = get_constraints(cib)

    # find all specified constrained resources and transform set options to
    # value pairs for normalization and validation
    resource_set_loaded_list = _load_resource_set_list(
        cib,
        env.report_processor,
        # dacite doesn't support TypedDicts (https://github.com/konradhalas/dacite/issues/125)
        # and therefore this command fails when called from APIv2, so we had to
        # use Mapping in the signature of the lib command
        cast(common.CmdInputResourceSetList, resource_set_list),
        validate.option_value_normalization(
            {
                "role": lambda value: value.capitalize(),
            }
        ),
    )
    # Unlike in plain constraints, validation cannot continue if even a single
    # resource could not be found. If such resources were omitted in their sets
    # for purposes of validation, similarly to plain constraint commands, then
    # those sets could become invalid, and thus validating such sets would
    # provide false results.
    if env.report_processor.has_errors:
        raise LibraryError()

    # transform constraint options to value pairs for normalization and
    # validation
    constraint_options_pairs = validate.values_to_pairs(
        constraint_options,
        validate.option_value_normalization(
            {
                "loss-policy": lambda value: value.lower(),
            }
        ),
    )

    # validation
    env.report_processor.report_list(
        ticket.validate_create_with_set(
            id_provider,
            resource_set_loaded_list,
            constraint_options_pairs,
            in_multiinstance_allowed=resource_in_clone_alowed,
        )
    )
    if env.report_processor.has_errors:
        raise LibraryError()

    # modify CIB
    new_constraint = ticket.create_with_set(
        constraint_section,
        id_provider,
        get_pacemaker_version_by_which_cib_was_validated(cib),
        _primitive_resource_set_list(resource_set_loaded_list),
        validate.pairs_to_values(constraint_options_pairs),
    )

    # Check whether the created constraint is a duplicate of an existing one
    env.report_processor.report_list(
        ticket.DuplicatesCheckerTicketWithSet().check(
            constraint_section,
            new_constraint,
            {reports.codes.FORCE} if duplication_alowed else set(),
        )
    )
    if env.report_processor.has_errors:
        raise LibraryError()

    # push CIB
    env.push_cib()


def remove(env: LibraryEnvironment, ticket_key: str, resource_id: str) -> bool:
    """
    remove all ticket constraint from resource
    If resource is in resource set with another resources then only resource
    ref is removed. If resource is alone in resource set whole constraint is
    removed.
    """
    constraint_section = get_constraints(env.get_cib())
    any_plain_removed = ticket.remove_plain(
        constraint_section, ticket_key, resource_id
    )
    any_with_resource_set_removed = ticket.remove_with_resource_set(
        constraint_section, ticket_key, resource_id
    )

    env.push_cib()

    return any_plain_removed or any_with_resource_set_removed



from collections.abc import Sequence
from typing import (
    Any,
    Final,
    Optional,
    Type,
    TypeVar,
    cast,
)

from lxml import etree
from lxml.etree import (
    _Attrib,
    _Element,
)

from pcs.common import reports
from pcs.common.fencing_topology import (
    TARGET_TYPE_ATTRIBUTE,
    TARGET_TYPE_NODE,
    TARGET_TYPE_REGEXP,
    FencingTargetType,
    FencingTargetValue,
)
from pcs.common.pacemaker.fencing_topology import (
    CibFencingLevelAttributeDto,
    CibFencingLevelNodeDto,
    CibFencingLevelRegexDto,
    CibFencingTopologyDto,
)
from pcs.common.reports import (
    ReportItemList,
    ReportItemSeverity,
    ReportProcessor,
    has_errors,
)
from pcs.common.reports import codes as report_codes
from pcs.common.reports.item import ReportItem
from pcs.common.types import StringSequence
from pcs.common.validate import is_integer
from pcs.lib.cib.const import TAG_FENCING_LEVEL
from pcs.lib.cib.resource.stonith import is_stonith
from pcs.lib.cib.tools import (
    ElementNotFound,
    IdProvider,
    get_element_by_id,
    get_fencing_topology,
    multivalue_attr_contains_value,
    multivalue_attr_delete_value,
    multivalue_attr_has_any_values,
)
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.state import _Element as StateElement
from pcs.lib.pacemaker.values import (
    sanitize_id,
    validate_id,
)

_DEVICES_ATTRIBUTE: Final = "devices"

FencingLevelDto = TypeVar(
    "FencingLevelDto",
    CibFencingLevelAttributeDto,
    CibFencingLevelNodeDto,
    CibFencingLevelRegexDto,
)


def _generate_level_id(
    id_provider: IdProvider,
    level: str,
    target_type: FencingTargetType,
    target_value: FencingTargetValue,
) -> str:
    if target_type == TARGET_TYPE_ATTRIBUTE:
        # For attribute target type, the value is tuple[str, str]
        id_part = target_value[0]
    else:
        # For all other target types, the value is str
        id_part = str(target_value)
    return id_provider.allocate_id(sanitize_id(f"fl-{id_part}-{level}"))


def add_level(  # noqa: PLR0913
    reporter: ReportProcessor,
    cib: _Element,
    level: str,
    target_type: FencingTargetType,
    target_value: FencingTargetValue,
    devices: StringSequence,
    cluster_status_nodes: Sequence[StateElement],
    level_id: Optional[str] = None,
    force_device: bool = False,
    force_node: bool = False,
) -> None:
    """
    Validate and add a new fencing level. Raise LibraryError if not valid.

    reporter -- report processor
    cib -- the whole cib
    level -- level (index) of the new fencing level
    target_type -- the new fencing level target value type
    target_value -- the new fencing level target value
    devices -- list of stonith devices for the new fencing level
    cluster_status_nodes -- list of status of existing cluster nodes
    level_id -- user specified id for the level element
    force_device -- continue even if a stonith device does not exist
    force_node -- continue even if a node (target) does not exist
    """
    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    id_provider = IdProvider(cib)
    validate_id_reports: ReportItemList = []
    if level_id is not None:
        validate_id(level_id, "fencing level id", reporter=validate_id_reports)
    reporter.report_list(
        (id_provider.book_ids(level_id) if level_id else [])
        + validate_id_reports
        + _validate_level(level)
        + _validate_target(
            cluster_status_nodes, target_type, target_value, force_node
        )
        + _validate_devices(cib, devices, force_device)
    )
    if reporter.has_errors:
        raise LibraryError()
    topology_el = get_fencing_topology(cib)
    reporter.report_list(
        _validate_level_target_devices_does_not_exist(
            topology_el, level, target_type, target_value, devices
        )
    )
    if reporter.has_errors:
        raise LibraryError()
    _append_level_element(
        topology_el,
        level,
        target_type,
        target_value,
        devices,
        (
            level_id
            or _generate_level_id(id_provider, level, target_type, target_value)
        ),
    )


def remove_all_levels(topology_el: _Element) -> None:
    """
    Remove all fencing levels.
    topology_el -- etree element to remove the levels from
    """
    # Do not ever remove a fencing-topology element, even if it is empty. There
    # may be ACLs set in pacemaker which allow "write" for fencing-level
    # elements (adding, changing and removing) but not fencing-topology
    # elements. In such a case, removing a fencing-topology element would cause
    # the whole change to be rejected by pacemaker with a "permission denied"
    # message.
    # https://bugzilla.redhat.com/show_bug.cgi?id=1642514
    for level_el in topology_el.findall(TAG_FENCING_LEVEL):
        # Parent is guaranteed by CIB schema
        cast(_Element, level_el.getparent()).remove(level_el)


def remove_levels_by_params(
    topology_el: _Element,
    level: Optional[str] = None,
    target_type: Optional[FencingTargetType] = None,
    target_value: Optional[FencingTargetValue] = None,
    devices: Optional[StringSequence] = None,
) -> ReportItemList:
    """
    Remove specified fencing level(s)

    topology_el -- etree element to remove the levels from
    level -- level (index) of the fencing level to remove
    target_type -- the removed fencing level target value type
    target_value -- the removed fencing level target value
    devices -- list of stonith devices of the removed fencing level
    """
    # Do not ever remove a fencing-topology element, even if it is empty. There
    # may be ACLs set in pacemaker which allow "write" for fencing-level
    # elements (adding, changing and removing) but not fencing-topology
    # elements. In such a case, removing a fencing-topology element would cause
    # the whole change to be rejected by pacemaker with a "permission denied"
    # message.
    # https://bugzilla.redhat.com/show_bug.cgi?id=1642514
    report_list: ReportItemList = []
    if target_type:
        report_list.extend(_validate_target_typewise(target_type))
        if has_errors(report_list):
            return report_list

    if devices is not None:
        for device_id in devices:
            validate_id(
                device_id, description="stonith id", reporter=report_list
            )

    level_el_list = _find_level_elements(
        topology_el, level, target_type, target_value, devices
    )

    if not level_el_list:
        report_list.append(
            ReportItem.error(
                reports.messages.CibFencingLevelDoesNotExist(
                    level or "",
                    target_type,
                    target_value,
                    sorted(devices) if devices else [],
                )
            )
        )
    if has_errors(report_list):
        return report_list
    for el in level_el_list:
        # Parent guaranteed by CIB schema
        cast(_Element, el.getparent()).remove(el)
    return report_list


def find_levels_with_device(
    topology_el: _Element, device_id: str
) -> list[_Element]:
    """
    Return list of all fencing-level elements that reference the specified
    device

    topology_el -- etree element with fencing levels
    device_id -- id of the stonith device
    """
    return [
        level_el
        for level_el in topology_el.findall(TAG_FENCING_LEVEL)
        if multivalue_attr_contains_value(
            level_el, _DEVICES_ATTRIBUTE, device_id
        )
    ]


def remove_device_from_level(level_el: _Element, device_id: str) -> None:
    """
    Remove specified stonith device from fencing level.

    level_el -- level element from which the device is removed
    device_id -- stonith device to remove
    """
    multivalue_attr_delete_value(level_el, _DEVICES_ATTRIBUTE, device_id)


def has_any_devices(level_el: _Element) -> bool:
    """
    Return whether there are any devices references in the fencing level

    level_el -- fencing level element
    """
    return multivalue_attr_has_any_values(level_el, _DEVICES_ATTRIBUTE)


# DEPRECATED, use fencing_topology_el_to_dto
def export(topology_el: _Element) -> list[dict[str, Any]]:
    """
    Export all fencing levels.

    Return a list of levels where each level is a dict with keys: target_type,
    target_value. level and devices. Devices is a list of stonith device ids.

    etree topology_el -- etree element to export
    """
    export_levels = []
    for level_el in topology_el.iterfind(TAG_FENCING_LEVEL):
        target_type: Optional[FencingTargetType] = None
        target_value: Optional[FencingTargetValue] = None
        if "target" in level_el.attrib:
            target_type = TARGET_TYPE_NODE
            target_value = level_el.get("target")
        elif "target-pattern" in level_el.attrib:
            target_type = TARGET_TYPE_REGEXP
            target_value = level_el.get("target-pattern")
        elif "target-attribute" in level_el.attrib:
            target_type = TARGET_TYPE_ATTRIBUTE
            target_value = (
                str(level_el.attrib["target-attribute"]),
                str(level_el.attrib["target-value"]),
            )
        if target_type and target_value:
            export_levels.append(
                {
                    "target_type": target_type,
                    "target_value": target_value,
                    "level": level_el.get("index"),
                    "devices": str(level_el.attrib[_DEVICES_ATTRIBUTE]).split(
                        ","
                    ),
                }
            )
    return export_levels


def _fencing_level_dto_factory(
    dto_type: Type[FencingLevelDto],
    level_el_attrib: _Attrib,
) -> FencingLevelDto:
    if dto_type is CibFencingLevelRegexDto:
        target_args = {"target_pattern": str(level_el_attrib["target-pattern"])}
    elif dto_type is CibFencingLevelAttributeDto:
        target_args = {
            "target_attribute": str(level_el_attrib["target-attribute"]),
            "target_value": str(level_el_attrib["target-value"]),
        }
    else:
        target_args = {"target": str(level_el_attrib["target"])}

    return dto_type(
        id=str(level_el_attrib["id"]),
        index=int(level_el_attrib["index"]),
        devices=[
            device.strip()
            for device in str(level_el_attrib["devices"]).split(",")
        ],
        **target_args,
    )


def fencing_topology_el_to_dto(
    fencing_topology_el: _Element,
) -> CibFencingTopologyDto:
    level_el_list = _find_all_level_elements(fencing_topology_el)
    target_node_list = []
    target_regex_list = []
    target_attr_list = []
    for level_el in level_el_list:
        if "target" in level_el.attrib:
            target_node_list.append(
                _fencing_level_dto_factory(
                    CibFencingLevelNodeDto,
                    level_el.attrib,
                )
            )
        if "target-pattern" in level_el.attrib:
            target_regex_list.append(
                _fencing_level_dto_factory(
                    CibFencingLevelRegexDto,
                    level_el.attrib,
                )
            )
        if "target-attribute" in level_el.attrib:
            target_attr_list.append(
                _fencing_level_dto_factory(
                    CibFencingLevelAttributeDto,
                    level_el.attrib,
                )
            )

    return CibFencingTopologyDto(
        target_node=target_node_list,
        target_regex=target_regex_list,
        target_attribute=target_attr_list,
    )


def verify(
    cib: _Element,
    cluster_status_nodes: Sequence[StateElement],
) -> ReportItemList:
    """
    Check if all cluster nodes and stonith devices used in fencing levels exist.

    cib -- the whole cib
    cluster_status_nodes -- list of status of existing cluster nodes
    """
    report_list: ReportItemList = []
    used_nodes: set[str] = set()
    used_devices: set[str] = set()
    topology_el = get_fencing_topology(cib)

    for level_el in topology_el.iterfind(TAG_FENCING_LEVEL):
        used_devices.update(
            str(level_el.get(_DEVICES_ATTRIBUTE, "")).split(",")
        )
        if "target" in level_el.attrib:
            used_nodes.add(str(level_el.get("target", "")))

    if used_devices:
        report_list.extend(
            _validate_devices(cib, sorted(used_devices), allow_force=False)
        )

    for node in sorted(used_nodes):
        report_list.extend(
            _validate_target_valuewise(
                cluster_status_nodes, TARGET_TYPE_NODE, node, allow_force=False
            )
        )
    return report_list


def _validate_level(level: str) -> ReportItemList:
    report_list: ReportItemList = []
    if not is_integer(level, 1, 9):
        report_list.append(
            ReportItem.error(
                reports.messages.InvalidOptionValue("level", str(level), "1..9")
            )
        )
    return report_list


def _validate_target(
    cluster_status_nodes: Sequence[StateElement],
    target_type: FencingTargetType,
    target_value: FencingTargetValue,
    force_node: bool = False,
) -> ReportItemList:
    return _validate_target_typewise(target_type) + _validate_target_valuewise(
        cluster_status_nodes, target_type, target_value, force_node
    )


def _validate_target_typewise(target_type: FencingTargetType) -> ReportItemList:
    report_list: ReportItemList = []
    if target_type not in [
        TARGET_TYPE_NODE,
        TARGET_TYPE_ATTRIBUTE,
        TARGET_TYPE_REGEXP,
    ]:
        report_list.append(
            ReportItem.error(
                reports.messages.InvalidOptionType(
                    "target",
                    ["node", "regular expression", "attribute_name=value"],
                )
            )
        )
    return report_list


def _validate_target_valuewise(
    cluster_status_nodes: Sequence[StateElement],
    target_type: FencingTargetType,
    target_value: FencingTargetValue,
    force_node: bool = False,
    allow_force: bool = True,
) -> ReportItemList:
    report_list: ReportItemList = []
    if target_type == TARGET_TYPE_NODE:
        node_found = False
        for node in cluster_status_nodes:
            if target_value == node.attrs.name:
                node_found = True
                break
        if not node_found:
            report_list.append(
                ReportItem(
                    severity=ReportItemSeverity(
                        level=(
                            ReportItemSeverity.WARNING
                            if force_node and allow_force
                            else ReportItemSeverity.ERROR
                        ),
                        force_code=(
                            None
                            if force_node or not allow_force
                            else report_codes.FORCE
                        ),
                    ),
                    message=reports.messages.NodeNotFound(
                        # This is a str based on target_type
                        str(target_value)
                    ),
                )
            )
    return report_list


def _validate_devices(
    cib: _Element,
    devices: StringSequence,
    force_device: bool = False,
    allow_force: bool = True,
) -> ReportItemList:
    report_list: ReportItemList = []
    if not devices:
        report_list.append(
            ReportItem.error(
                reports.messages.RequiredOptionsAreMissing(["stonith devices"])
            )
        )
    invalid_devices = []
    for dev in devices:
        validate_id_report_list: ReportItemList = []
        validate_id(
            dev, description="device id", reporter=validate_id_report_list
        )
        report_list.extend(validate_id_report_list)
        if has_errors(validate_id_report_list):
            continue
        try:
            if not is_stonith(get_element_by_id(cib, dev)):
                invalid_devices.append(dev)
        except ElementNotFound:
            invalid_devices.append(dev)
    if invalid_devices:
        report_list.append(
            ReportItem(
                severity=ReportItemSeverity(
                    level=(
                        ReportItemSeverity.WARNING
                        if force_device and allow_force
                        else ReportItemSeverity.ERROR
                    ),
                    force_code=(
                        None
                        if force_device or not allow_force
                        else report_codes.FORCE
                    ),
                ),
                message=reports.messages.StonithResourcesDoNotExist(
                    invalid_devices
                ),
            )
        )
    return report_list


def _validate_level_target_devices_does_not_exist(
    tree: _Element,
    level: str,
    target_type: FencingTargetType,
    target_value: FencingTargetValue,
    devices: StringSequence,
) -> ReportItemList:
    report_list: ReportItemList = []
    if _find_level_elements(tree, level, target_type, target_value, devices):
        report_list.append(
            ReportItem.error(
                reports.messages.CibFencingLevelAlreadyExists(
                    level, target_type, target_value, list(devices)
                )
            )
        )
    return report_list


def _append_level_element(
    tree: _Element,
    level: str,
    target_type: FencingTargetType,
    target_value: FencingTargetValue,
    devices: StringSequence,
    level_id: str,
) -> _Element:
    level_el = etree.SubElement(
        tree,
        TAG_FENCING_LEVEL,
        index=level,
        devices=",".join(devices),
        id=level_id,
    )
    if target_type == TARGET_TYPE_NODE:
        level_el.set("target", str(target_value))
    elif target_type == TARGET_TYPE_REGEXP:
        level_el.set("target-pattern", str(target_value))
    elif target_type == TARGET_TYPE_ATTRIBUTE:
        level_el.set("target-attribute", target_value[0])
        level_el.set("target-value", target_value[1])
    return level_el


def _find_all_level_elements(tree: _Element) -> list[_Element]:
    return tree.findall(TAG_FENCING_LEVEL)


def _find_level_elements(
    tree: _Element,
    level: Optional[str] = None,
    target_type: Optional[FencingTargetType] = None,
    target_value: Optional[FencingTargetValue] = None,
    devices: Optional[StringSequence] = None,
) -> list[_Element]:
    xpath_vars: dict[str, str] = {}

    xpath_target = ""
    if target_type and target_value:
        if target_type == TARGET_TYPE_NODE:
            xpath_target = "@target=$var_target"
            # The value of xpath_target determines that this is a string
            xpath_vars["var_target"] = str(target_value)
        elif target_type == TARGET_TYPE_REGEXP:
            xpath_target = "@target-pattern=$var_target_pattern"
            # The value of xpath_target determines that this is a string
            xpath_vars["var_target_pattern"] = str(target_value)
        elif target_type == TARGET_TYPE_ATTRIBUTE:
            xpath_target = (
                "@target-attribute=$var_target_attribute "
                "and @target-value=$var_target_value"
            )
            xpath_vars["var_target_attribute"] = target_value[0]
            xpath_vars["var_target_value"] = target_value[1]
    xpath_devices = ""
    if devices:
        xpath_devices = "@devices=$var_devices"
    xpath_level = ""
    if level:
        xpath_level = "@index=$var_level"

    xpath_attrs = " and ".join(
        filter(None, [xpath_level, xpath_devices, xpath_target])
    )
    if xpath_attrs:
        return tree.xpath(
            f"{TAG_FENCING_LEVEL}[{xpath_attrs}]",
            var_devices=(",".join(devices) if devices else ""),
            var_level=level or "",
            **xpath_vars,  # type: ignore
        )

    return _find_all_level_elements(tree)



from typing import Mapping

from pcs.common import pcs_pycurl as pycurl
from pcs.common import reports
from pcs.common.host import PcsKnownHost
from pcs.common.node_communicator import (
    HostNotFound,
    NodeTargetFactory,
    RequestTarget,
)
from pcs.common.reports import (
    ReportItemSeverity,
    ReportProcessor,
)
from pcs.common.reports.item import ReportItem, ReportItemList
from pcs.common.types import StringIterable
from pcs.lib.errors import LibraryError


class NodeTargetLibFactory(NodeTargetFactory):
    def __init__(
        self,
        known_hosts: Mapping[str, PcsKnownHost],
        report_processor: ReportProcessor,
    ):
        super().__init__(known_hosts)
        self._report_processor = report_processor

    def get_target_list_with_reports(
        self,
        host_name_list: StringIterable,
        skip_non_existing: bool = False,
        allow_skip: bool = True,
        report_none_host_found: bool = True,
    ) -> tuple[ReportItemList, list[RequestTarget]]:
        target_list = []
        unknown_host_list = []
        for host_name in host_name_list:
            try:
                target_list.append(self.get_target(host_name))
            except HostNotFound:
                unknown_host_list.append(host_name)

        report_list = []
        if unknown_host_list:
            report_list.append(
                ReportItem(
                    severity=reports.item.get_severity(
                        (
                            reports.codes.SKIP_OFFLINE_NODES
                            if allow_skip
                            else None
                        ),
                        skip_non_existing,
                    ),
                    message=reports.messages.HostNotFound(
                        sorted(unknown_host_list),
                    ),
                )
            )

        if not target_list and host_name_list and report_none_host_found:
            # we want to create this report only if there was at least one
            # required address specified
            report_list.append(
                ReportItem.error(reports.messages.NoneHostFound())
            )
        return report_list, target_list

    def get_target_list(
        self, host_name_list, skip_non_existing=False, allow_skip=True
    ):
        report_list, target_list = self.get_target_list_with_reports(
            host_name_list, skip_non_existing, allow_skip
        )
        if report_list:
            if self._report_processor.report_list(report_list).has_errors:
                raise LibraryError()
        return target_list


def response_to_report_item(
    response,
    severity=ReportItemSeverity.ERROR,
    forceable=None,
    report_pcsd_too_old_on_404=False,
):
    """
    Returns report item which corresponds to response if was not successful.
    Otherwise returns None.

    Response response -- response from which report item shoculd be created
    ReportItemseverity severity -- severity of report item
    string forceable -- force code
    bool report_pcsd_too_old_on_404 -- if False, report unsupported command
    """
    response_code = response.response_code
    report_item = None
    reason = None
    if (
        report_pcsd_too_old_on_404
        and response.was_connected
        and response_code == 404
    ):
        return ReportItem.error(
            reports.messages.PcsdVersionTooOld(response.request.host_label)
        )
    if response.was_connected:
        if response_code == 400:
            # old pcsd protocol: error messages are commonly passed in plain
            # text in response body with HTTP code 400
            # we need to be backward compatible with that
            report_item = reports.messages.NodeCommunicationCommandUnsuccessful
            reason = response.data.rstrip()
        elif response_code == 401:
            report_item = reports.messages.NodeCommunicationErrorNotAuthorized
            reason = f"HTTP error: {response_code}"
        elif response_code == 403:
            report_item = (
                reports.messages.NodeCommunicationErrorPermissionDenied
            )
            reason = f"HTTP error: {response_code}"
        elif response_code == 404:
            report_item = (
                reports.messages.NodeCommunicationErrorUnsupportedCommand
            )
            reason = f"HTTP error: {response_code}"
        elif response_code >= 400:
            report_item = reports.messages.NodeCommunicationError
            reason = f"HTTP error: {response_code}"
    elif response.errno in [
        pycurl.E_OPERATION_TIMEDOUT,
        pycurl.E_OPERATION_TIMEOUTED,
    ]:
        report_item = reports.messages.NodeCommunicationErrorTimedOut
        reason = response.error_msg
    else:
        report_item = reports.messages.NodeCommunicationErrorUnableToConnect
        reason = response.error_msg
    if not report_item:
        return None
    return ReportItem(
        severity=ReportItemSeverity(severity, forceable),
        message=report_item(
            response.request.host_label,
            response.request.action,
            reason,
        ),
    )



import io
from unittest import TestCase

import pcs.lib.node_communication as lib
from pcs.common import pcs_pycurl as pycurl
from pcs.common.host import (
    Destination,
    PcsKnownHost,
)
from pcs.common.node_communicator import (
    Request,
    RequestData,
    RequestTarget,
    Response,
)
from pcs.common.reports import ReportItemSeverity as severity
from pcs.common.reports import codes as report_codes

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_report_item_equal,
)
from pcs_test.tools.custom_mock import (
    MockCurl,
    MockLibraryReportProcessor,
)


class NodeTargetLibFactory(TestCase):
    def setUp(self):
        self.known_hosts = {
            "host{}".format(i): PcsKnownHost(
                "host{}".format(i),
                "token{}".format(i),
                [
                    Destination(
                        "addr{}{}".format(i, j), "port{}{}".format(i, j)
                    )
                    for j in range(2)
                ],
            )
            for i in range(2)
        }
        self.report_processor = MockLibraryReportProcessor()
        self.factory = lib.NodeTargetLibFactory(
            self.known_hosts, self.report_processor
        )

    def assert_equal_known_host_target(self, known_host, target):
        self.assertEqual(known_host.name, target.label)
        self.assertEqual(known_host.token, target.token)
        self.assertEqual(known_host.dest_list, target.dest_list)

    def test_one_host(self):
        host = "host0"
        self.assert_equal_known_host_target(
            self.known_hosts[host], self.factory.get_target_list([host])[0]
        )
        self.report_processor.assert_reports([])

    def test_multiple_hosts(self):
        host_list = ["host0", "host1"]
        target_list = self.factory.get_target_list(host_list)
        for i, host in enumerate(host_list):
            self.assert_equal_known_host_target(
                self.known_hosts[host], target_list[i]
            )
        self.report_processor.assert_reports([])

    def test_multiple_not_found(self):
        host = "host0"
        unknown_hosts = ["node0", "node1"]
        report = fixture.error(
            report_codes.HOST_NOT_FOUND,
            force_code=report_codes.SKIP_OFFLINE_NODES,
            host_list=unknown_hosts,
        )
        assert_raise_library_error(
            lambda: self.factory.get_target_list([host] + unknown_hosts)
        )
        self.report_processor.assert_reports([report])

    def test_multiple_skip_not_allowed(self):
        host = "host0"
        unknown_hosts = ["node0", "node1"]
        report = fixture.error(
            report_codes.HOST_NOT_FOUND, host_list=unknown_hosts
        )
        assert_raise_library_error(
            lambda: self.factory.get_target_list(
                [host] + unknown_hosts,
                allow_skip=False,
            )
        )
        self.report_processor.assert_reports([report])

    def test_multiple_not_found_skip_offline(self):
        host = "host0"
        unknown_hosts = ["node0", "node1"]
        target_list = self.factory.get_target_list(
            [host] + unknown_hosts, skip_non_existing=True
        )
        self.assert_equal_known_host_target(
            self.known_hosts[host], target_list[0]
        )
        self.report_processor.assert_reports(
            [fixture.warn(report_codes.HOST_NOT_FOUND, host_list=unknown_hosts)]
        )

    def test_no_host_found(self):
        unknown_hosts = ["node0", "node1"]
        report_list = [
            fixture.error(
                report_codes.HOST_NOT_FOUND,
                force_code=report_codes.SKIP_OFFLINE_NODES,
                host_list=unknown_hosts,
            ),
            fixture.error(report_codes.NONE_HOST_FOUND),
        ]
        assert_raise_library_error(
            lambda: self.factory.get_target_list(unknown_hosts),
        )
        self.report_processor.assert_reports(report_list)

    def test_no_host_found_skip_offline(self):
        unknown_hosts = ["node0", "node1"]
        report_list = [
            fixture.warn(report_codes.HOST_NOT_FOUND, host_list=unknown_hosts),
            fixture.error(report_codes.NONE_HOST_FOUND),
        ]
        assert_raise_library_error(
            lambda: self.factory.get_target_list(
                unknown_hosts, skip_non_existing=True
            )
        )
        self.report_processor.assert_reports(report_list)

    def test_empty_host_list(self):
        self.assertEqual([], self.factory.get_target_list([]))
        self.report_processor.assert_reports([])


class ResponseToReportItemTest(TestCase):
    def fixture_response_connected(self, response_code):
        handle = MockCurl({pycurl.RESPONSE_CODE: response_code})
        handle.request_obj = Request(
            RequestTarget(self.host), RequestData(self.request)
        )
        handle.output_buffer = io.BytesIO()
        handle.output_buffer.write(self.data)
        return Response.connection_successful(handle)

    def fixture_response_not_connected(self, errno, error_msg):
        handle = MockCurl()
        handle.request_obj = Request(
            RequestTarget(self.host), RequestData(self.request)
        )
        return Response.connection_failure(handle, errno, error_msg)

    def setUp(self):
        self.host = "host"
        self.request = "request"
        self.data = b"data"

    def test_code_200(self):
        self.assertIsNone(
            lib.response_to_report_item(self.fixture_response_connected(200))
        )

    def test_code_400(self):
        assert_report_item_equal(
            lib.response_to_report_item(self.fixture_response_connected(400)),
            (
                severity.ERROR,
                report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                {
                    "node": self.host,
                    "command": self.request,
                    "reason": self.data.decode("utf-8"),
                },
                None,
            ),
        )

    def test_code_401(self):
        assert_report_item_equal(
            lib.response_to_report_item(self.fixture_response_connected(401)),
            (
                severity.ERROR,
                report_codes.NODE_COMMUNICATION_ERROR_NOT_AUTHORIZED,
                {
                    "node": self.host,
                    "command": self.request,
                    "reason": "HTTP error: 401",
                },
                None,
            ),
        )

    def test_code_403(self):
        assert_report_item_equal(
            lib.response_to_report_item(self.fixture_response_connected(403)),
            (
                severity.ERROR,
                report_codes.NODE_COMMUNICATION_ERROR_PERMISSION_DENIED,
                {
                    "node": self.host,
                    "command": self.request,
                    "reason": "HTTP error: 403",
                },
                None,
            ),
        )

    def test_code_404(self):
        assert_report_item_equal(
            lib.response_to_report_item(self.fixture_response_connected(404)),
            (
                severity.ERROR,
                report_codes.NODE_COMMUNICATION_ERROR_UNSUPPORTED_COMMAND,
                {
                    "node": self.host,
                    "command": self.request,
                    "reason": "HTTP error: 404",
                },
                None,
            ),
        )

    def test_code_other(self):
        assert_report_item_equal(
            lib.response_to_report_item(self.fixture_response_connected(500)),
            (
                severity.ERROR,
                report_codes.NODE_COMMUNICATION_ERROR,
                {
                    "node": self.host,
                    "command": self.request,
                    "reason": "HTTP error: 500",
                },
                None,
            ),
        )

    def test_timed_out(self):
        response = self.fixture_response_not_connected(
            pycurl.E_OPERATION_TIMEDOUT, "err"
        )
        assert_report_item_equal(
            lib.response_to_report_item(response),
            (
                severity.ERROR,
                report_codes.NODE_COMMUNICATION_ERROR_TIMED_OUT,
                {"node": self.host, "command": self.request, "reason": "err"},
                None,
            ),
        )

    def test_timeouted(self):
        response = self.fixture_response_not_connected(
            pycurl.E_OPERATION_TIMEOUTED, "err"
        )
        assert_report_item_equal(
            lib.response_to_report_item(response),
            (
                severity.ERROR,
                report_codes.NODE_COMMUNICATION_ERROR_TIMED_OUT,
                {"node": self.host, "command": self.request, "reason": "err"},
                None,
            ),
        )

    def test_unable_to_connect(self):
        response = self.fixture_response_not_connected(
            pycurl.E_SEND_ERROR, "err"
        )
        assert_report_item_equal(
            lib.response_to_report_item(response),
            (
                severity.ERROR,
                report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                {"node": self.host, "command": self.request, "reason": "err"},
                None,
            ),
        )
